// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "bitmanip.h"
#include "twiddles.h"
#include "fft_r2_dpfp.h"

// ----------------------------------------------------------------
void fft_r2_dpfp(
	complex_double * y,
	fft_opts_t       opts,
	fft_verbosity_t  vopts)
{
	int N = opts.N;
	int log2N = splog2((unsigned)N);

	if (!is_a_power_of_two(N)) {
		std::cerr << "fft_r2_dpfp: input length "
			<< N << " is not a power of two.\n";
		exit(1);
	}

	tvector<complex_double> W = half_twiddle_table_double(N, opts.forward);
	complex_double   a, b, w, wb, c, d;
	int num_stages;
	int stageno;
	int i;
	int num_flocks;
	int flock_sep;
	int bperflock;
	int wingspan;
	int twiddle_stride;
	int fno;
	int bno;
	int upperidx;
	int loweridx;
	int twiddleidx;
	double sqrt_recip_2 = sqrt(0.5);

	num_stages = log2N;

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Output folding.
	if (opts.fold_out) {
		for (i = 1; i < N; i+=2)
			y[i] = -y[i];
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Bit-reverse stage.
	for (i = 0; i < N; i++) {
		int ir = bit_reverse(i, log2N);
		complex_double temp;
		if (i < ir) {
			temp  = y[i];
			y[i]  = y[ir];
			y[ir] = temp;
		}
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Radix-two stages.
	//
	// y00 \ / @    \ / @    \ / @    \ / Y00
	// y08 / \ @    \ / @    \ / @    \ / Y01
	// y04 \ / @ W0 / \ @    \ / @    \ / Y02
	// y12 / \ @ W4 / \ @    \ / @    \ / Y03
	//
	// y02 \ / @    \ / @ W0 / \ @    \ / Y04
	// y10 / \ @    \ / @ W2 / \ @    \ / Y05
	// y06 \ / @ W0 / \ @ W4 / \ @    \ / Y06
	// y14 / \ @ W4 / \ @ W6 / \ @    \ / Y07
	//
	// y01 \ / @    \ / @    \ / @ W0 / \ Y08
	// y09 / \ @    \ / @    \ / @ W1 / \ Y09
	// y05 \ / @ W0 / \ @    \ / @ W2 / \ Y10
	// y13 / \ @ W4 / \ @    \ / @ W3 / \ Y11
	//
	// y03 \ / @    \ / @ W0 / \ @ W4 / \ Y12
	// y11 / \ @    \ / @ W2 / \ @ W5 / \ Y13
	// y07 \ / @ W0 / \ @ W4 / \ @ W6 / \ Y14
	// y15 / \ @ W4 / \ @ W6 / \ @ W7 / \ Y15
	//
	// a   \ / c   --> c := a + b
	// b   / \ d   --> d := a - b
	//
	// a   \ / c   --> c := a + w * b
	// b w / \ d   --> d := a - w * b
	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	num_flocks = N/2;
	flock_sep  = 2;
	bperflock  = 1;
	wingspan   = 1;
	twiddle_stride = N/2;

	for (stageno = 0; stageno < num_stages; stageno++) {

		if (vopts.verbose) {
			printf("\n");
			printf("Host r2 dpfp stage %d\n", stageno);
		}

		for (fno = 0; fno < num_flocks; fno++) {
			twiddleidx = 0;

			for (bno = 0; bno < bperflock; bno++) {
				upperidx = fno * flock_sep + bno;
				loweridx = upperidx + wingspan;

				a  = y[upperidx];
				b  = y[loweridx];

				w  = W[twiddleidx];
				wb = w * b;
				c  = a + wb;
				d  = a - wb;

				if (vopts.verbose) {
					printf("H %3d %3d  %3d\n",
						upperidx, loweridx, twiddleidx);
				}

				if (vopts.extra_verbose) {
					printf("HT w=%15.10lf,%15.10lf\n",
						w.re, w.im);
					printf("HI a=%15.10lf,%15.10lf b=%15.10lf,%15.10lf\n",
						a.re, a.im, b.re, b.im);
				}

				if (opts.scale) {
					c *= sqrt_recip_2;
					d *= sqrt_recip_2;
				}

				y[upperidx] = c;
				y[loweridx] = d;

				if (vopts.extra_verbose) {
					printf("HO a=%15.10lf,%15.10lf b=%15.10lf,%15.10lf\n",
						c.re, c.im, d.re, d.im);
				}

				twiddleidx += twiddle_stride;
			}
		}

		num_flocks >>= 1;
		flock_sep  <<= 1;
		bperflock  <<= 1;
		wingspan   <<= 1;
		twiddle_stride >>= 1;
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Input folding.

	if (opts.fold_in) {
		for (i = 1; i < N; i+=2)
			y[i] = -y[i];
	}
}
