// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "tvector.h"
#include "complex_float.h"
#include "complex_double.h"
#include "comparef.h"
#include "compared.h"
#include "comparecf.h"
#include "comparecd.h"
#include "nbio.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {-rs|-rd|-cs|-cd} {input file 1} {input file 2}\n",
		argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	int    argb = 1;
	char * input_file_name_1;
	char * input_file_name_2;
	int    do_complex = 0;
	int    do_double  = 0;
	int    show_file_names = 0;
	float  tol = 1e-6;
	int    rc  = 1;

	if (argc < 2)
		usage(argv[0]);
	while ((argb < argc) && (argv[argb][0] == '-')) {
		if (strcmp(argv[argb], "-rs") == 0) {
			do_complex = 0;
			do_double  = 0;
		}
		else if (strcmp(argv[argb], "-rd") == 0) {
			do_complex = 0;
			do_double  = 1;
		}
		else if (strcmp(argv[argb], "-cs") == 0) {
			do_complex = 1;
			do_double  = 0;
		}
		else if (strcmp(argv[argb], "-cd") == 0) {
			do_complex = 1;
			do_double  = 1;
		}
		else if (strcmp(argv[argb], "-sfn") == 0) {
			show_file_names = 1;
		}
		else if (strcmp(argv[argb], "-tol") == 0) {
			if ((argc - argb) < 2)
				usage(argv[0]);
			argb++;
			if (sscanf(argv[argb], "%f", &tol) != 1)
				usage(argv[0]);
		}
		else {
			usage(argv[0]);
		}
		argb++;
	}
	if ((argc - argb) != 2)
		usage(argv[0]);
	input_file_name_1 = argv[argb];
	input_file_name_2 = argv[argb + 1];

	if (show_file_names)
		printf("%s vs. %s:\n", input_file_name_1, input_file_name_2);

	if (do_complex) {
		if (do_double) {
			tvector<complex_double> in1, in2, diff;
			double  maxrelerr, rmsrelerr, maxabserr, maxabsval;
			load_from_binary_file_or_die_nbo(in1,input_file_name_1);
			load_from_binary_file_or_die_nbo(in2, input_file_name_2);
			(void)complex_double_vector_compare(in1.expose(), in2.expose(),
				in1.get_num_elements(), tol, maxrelerr, rmsrelerr,
				maxabserr, maxabsval);
			printf("Max relative error %.6le\n", maxrelerr);
			printf("RMS relative error %.6le\n", rmsrelerr);
			printf("Max absolute error %.6le\n", maxabserr);
			printf("Max absolute value %.6le\n", maxabsval);
			rc = (rmsrelerr <= tol) ? 0 : 1;
		}
		else {
			tvector<complex_float> in1, in2, diff;
			float  maxrelerr, rmsrelerr, maxabserr, maxabsval;
			load_from_binary_file_or_die_nbo(in1,input_file_name_1);
			load_from_binary_file_or_die_nbo(in2, input_file_name_2);
			(void)complex_float_vector_compare(in1.expose(), in2.expose(),
				in1.get_num_elements(), tol, maxrelerr, rmsrelerr,
				maxabserr, maxabsval);
			printf("Max relative error %.6e\n", maxrelerr);
			printf("RMS relative error %.6e\n", rmsrelerr);
			printf("Max absolute error %.6e\n", maxabserr);
			printf("Max absolute value %.6e\n", maxabsval);
			rc = (rmsrelerr <= tol) ? 0 : 1;
		}
	}
	else {
		if (do_double) {
			tvector<double> in1, in2, diff;
			double  maxrelerr, rmsrelerr, maxabserr, maxabsval;
			load_from_binary_file_or_die_nbo(in1, input_file_name_1);
			load_from_binary_file_or_die_nbo(in2, input_file_name_2);
			(void)double_vector_compare(in1.expose(), in2.expose(),
				in1.get_num_elements(), tol, maxrelerr, rmsrelerr,
				maxabserr, maxabsval);
			printf("Max relative error %.6le\n", maxrelerr);
			printf("RMS relative error %.6le\n", rmsrelerr);
			printf("Max absolute error %.6le\n", maxabserr);
			printf("Max absolute value %.6le\n", maxabsval);
			rc = (rmsrelerr <= tol) ? 0 : 1;
		}
		else {
			tvector<float> in1, in2, diff;
			float  maxrelerr, rmsrelerr, maxabserr, maxabsval;
			load_from_binary_file_or_die_nbo(in1, input_file_name_1);
			load_from_binary_file_or_die_nbo(in2, input_file_name_2);
			(void)float_vector_compare(in1.expose(), in2.expose(),
				in1.get_num_elements(), tol, maxrelerr, rmsrelerr,
				maxabserr, maxabsval);

			int nleft = 0;
			int nright = 0;
			int nat = 0;
			int n = in1.get_num_elements();
			float * p1 = in1.expose();
			float * p2 = in2.expose();
			for (int i = 0; i < n; i++) {
				float diff = p1[i] - p2[i];
				if (diff > 0)
					nleft++;
				else if (diff < 0)
					nright++;
				else
					nat++;
			}
			printf("Max relative error %.6e\n", maxrelerr);
			printf("RMS relative error %.6e\n", rmsrelerr);
			printf("Max absolute error %.6e\n", maxabserr);
			printf("Max absolute value %.6e\n", maxabsval);
			//printf("Balance %d/%d/%d\n",
			printf("Balance %d < %d > %d [%d]\n",
				(int)((100.0 * nleft)  / n),
				(int)((100.0 * nat)    / n),
				(int)((100.0 * nright) / n),
				n);

			rc = (rmsrelerr <= tol) ? 0 : 1;
		}
	}
	if (show_file_names)
		printf("\n");

	return rc;
}
