// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include "tvector.h"
#include "complex_float.h"
#include "complex_double.h"
#include "vector_print.h"
#include "nbio.h"

#define FMT_BIN      20
#define FMT_TXT      21
#define FMT_HEX      22

#define PREC_SINGLE  30
#define PREC_DOUBLE  31

#define CPLX_RECT    40
#define CPLX_RE      41
#define CPLX_IM      42
#define CPLX_POLAR   43
#define CPLX_MAG     44
#define CPLX_MAG_LOG 45
#define CPLX_PHZ     46

struct opts_t {
	char * argv0;
	int    input_precision;
	int    input_file_format;
	int    input_cplx_format;
	char * input_file_name;
	int    output_precision;
	int    output_file_format;
	int    output_cplx_format;
	char * output_file_name;
};

static void set_defaults(
	opts_t & ropts);

static int parse_command_line(
	opts_t & ropts,
	int      argc,
	char  ** argv);

static void coding_error(
	char * file,
	int    line);

static void single_to_single(
	opts_t opts);
static void convert(
	opts_t opts);

#define CODING_ERROR() coding_error(__FILE__, __LINE__)

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s [options] {input file} {output file}\n",
		argv0);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -ib|-ob|-b:    Binary format for input, output, or both.\n");
	fprintf(stderr, "  -it|-ot|-t:    Text format for input, output, or both.\n");
	fprintf(stderr, "  -ix|-ox|-x:    Hex  format for input, output, or both.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "  -is|-os|-s:    Single precision for input, output, or both.\n");
	fprintf(stderr, "  -id|-od|-d:    Double precision for input, output, or both.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "  -irect|-orect|-rect: Rectangular format for input, output, or both.\n");
	fprintf(stderr, "  -ire|-ore|-re: Real-only format for input, output, or both.\n");
	fprintf(stderr, "  -iim|-oim|-im: Imag-only format for input, output, or both.\n");
	fprintf(stderr, "  -ipol|-opol|-pol: Polar format for input, output, or both.\n");
	fprintf(stderr, "  -imag|-omag|-mag: Magnitude only for input, output, or both.\n");
	fprintf(stderr, "  -iphz|-ophz|-phz: Phase only for input, output, or both.\n");
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	opts_t opts;
	set_defaults(opts);
	if (!parse_command_line(opts, argc, argv))
		usage(argv[0]);

	if ((opts.input_precision == PREC_SINGLE)
		&& (opts.output_precision == PREC_SINGLE))
	{
		single_to_single(opts);
	}
	else {
		convert(opts);
	}

	return 0;
}

// ----------------------------------------------------------------
static void set_defaults(
	opts_t & ropts)
{
	ropts.input_precision    = PREC_SINGLE;
	ropts.output_precision   = PREC_SINGLE;
	ropts.input_file_format  = FMT_BIN;
	ropts.output_file_format = FMT_BIN;
	ropts.input_cplx_format  = CPLX_RECT;
	ropts.output_cplx_format = CPLX_RECT;
}

// ----------------------------------------------------------------
static int parse_command_line(
	opts_t & ropts,
	int      argc,
	char  ** argv)
{
	if (argc < 2)
		return 0;
	int argi = 1;
	ropts.argv0 = argv[0];
	if (argc < 2)
		return 0;
	while (argv[argi][0] == '-') {
		if (strcmp(argv[argi], "--help") == 0)
			return 0;

		else if (strcmp(argv[argi], "-s") == 0)
			ropts.input_precision = ropts.output_precision = PREC_SINGLE;
		else if (strcmp(argv[argi], "-is") == 0)
			ropts.input_precision = PREC_SINGLE;
		else if (strcmp(argv[argi], "-os") == 0)
			ropts.output_precision = PREC_SINGLE;

		else if (strcmp(argv[argi], "-d") == 0)
			ropts.input_precision = ropts.output_precision = PREC_DOUBLE;
		else if (strcmp(argv[argi], "-id") == 0)
			ropts.input_precision = PREC_DOUBLE;
		else if (strcmp(argv[argi], "-od") == 0)
			ropts.output_precision = PREC_DOUBLE;


		else if (strcmp(argv[argi], "-b") == 0)
			ropts.input_file_format = ropts.output_file_format = FMT_BIN;
		else if (strcmp(argv[argi], "-ib") == 0)
			ropts.input_file_format = FMT_BIN;
		else if (strcmp(argv[argi], "-ob") == 0)
			ropts.output_file_format = FMT_BIN;

		else if (strcmp(argv[argi], "-t") == 0)
			ropts.input_file_format = ropts.output_file_format = FMT_TXT;
		else if (strcmp(argv[argi], "-it") == 0)
			ropts.input_file_format = FMT_TXT;
		else if (strcmp(argv[argi], "-ot") == 0)
			ropts.output_file_format = FMT_TXT;

		else if (strcmp(argv[argi], "-x") == 0)
			ropts.input_file_format = ropts.output_file_format = FMT_HEX;
		else if (strcmp(argv[argi], "-ix") == 0)
			ropts.input_file_format = FMT_HEX;
		else if (strcmp(argv[argi], "-ox") == 0)
			ropts.output_file_format = FMT_HEX;


		else if (strcmp(argv[argi], "-rect") == 0)
			ropts.input_cplx_format = ropts.output_cplx_format = CPLX_RECT;
		else if (strcmp(argv[argi], "-irect") == 0)
			ropts.input_cplx_format = CPLX_RECT;
		else if (strcmp(argv[argi], "-orect") == 0)
			ropts.output_cplx_format = CPLX_RECT;

		else if (strcmp(argv[argi], "-re") == 0)
			ropts.input_cplx_format = ropts.output_cplx_format = CPLX_RE;
		else if (strcmp(argv[argi], "-ire") == 0)
			ropts.input_cplx_format = CPLX_RE;
		else if (strcmp(argv[argi], "-ore") == 0)
			ropts.output_cplx_format = CPLX_RE;

		else if (strcmp(argv[argi], "-im") == 0)
			ropts.input_cplx_format = ropts.output_cplx_format = CPLX_IM;
		else if (strcmp(argv[argi], "-iim") == 0)
			ropts.input_cplx_format = CPLX_IM;
		else if (strcmp(argv[argi], "-oim") == 0)
			ropts.output_cplx_format = CPLX_IM;

		else if (strcmp(argv[argi], "-pol") == 0)
			ropts.input_cplx_format = ropts.output_cplx_format = CPLX_POLAR;
		else if (strcmp(argv[argi], "-ipol") == 0)
			ropts.input_cplx_format = CPLX_POLAR;
		else if (strcmp(argv[argi], "-opol") == 0)
			ropts.output_cplx_format = CPLX_POLAR;

		else if (strcmp(argv[argi], "-mag") == 0)
			ropts.input_cplx_format = ropts.output_cplx_format = CPLX_MAG;
		else if (strcmp(argv[argi], "-imag") == 0)
			ropts.input_cplx_format = CPLX_MAG;
		else if (strcmp(argv[argi], "-omag") == 0)
			ropts.output_cplx_format = CPLX_MAG;

		else if (strcmp(argv[argi], "-maglog") == 0)
			ropts.input_cplx_format = ropts.output_cplx_format = CPLX_MAG_LOG;
		else if (strcmp(argv[argi], "-imaglog") == 0)
			ropts.input_cplx_format = CPLX_MAG_LOG;
		else if (strcmp(argv[argi], "-omaglog") == 0)
			ropts.output_cplx_format = CPLX_MAG_LOG;

		else if (strcmp(argv[argi], "-phz") == 0)
			ropts.input_cplx_format = ropts.output_cplx_format = CPLX_PHZ;
		else if (strcmp(argv[argi], "-iphz") == 0)
			ropts.input_cplx_format = CPLX_PHZ;
		else if (strcmp(argv[argi], "-ophz") == 0)
			ropts.output_cplx_format = CPLX_PHZ;

		else
			return 0;

		argi++;

		if (argi >= argc)
			break;
	}

	if ((argc - argi) != 2)
		return 0;
	ropts.input_file_name  = argv[argi];
	ropts.output_file_name = argv[argi+1];
	return 1;
}

// ----------------------------------------------------------------
static void coding_error(
	char * file,
	int    line)
{
	fprintf(stderr, "Coding error detected in file %s, line %d\n",
		file, line);
	exit(1);
}

// ----------------------------------------------------------------
static void single_to_single(
	opts_t opts)
{
	tvector<complex_float> csdata;
	tvector<float> rsdata;
	int i, n = 0;

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Read input data
	switch (opts.input_file_format) {

	case FMT_BIN:
		switch (opts.input_cplx_format) {
		case CPLX_RE:
		case CPLX_IM:
		case CPLX_MAG:
		case CPLX_MAG_LOG:
		case CPLX_PHZ:
			load_from_binary_file_or_die_nbo(rsdata, opts.input_file_name);
			n = rsdata.get_num_elements();
			break;
		case CPLX_RECT:
		case CPLX_POLAR:
			load_from_binary_file_or_die_nbo(csdata, opts.input_file_name);
			n = csdata.get_num_elements();
			break;
		default:
			CODING_ERROR();
			break;
		}
		break;

	case FMT_TXT:
		switch (opts.input_cplx_format) {
		case CPLX_RE:
		case CPLX_IM:
		case CPLX_MAG:
		case CPLX_MAG_LOG:
		case CPLX_PHZ:
			if (!rsdata.load_from_text_file(opts.input_file_name)) {
				fprintf(stderr, "%s:  couldn't read \"%s\".\n",
					opts.argv0, opts.input_file_name);
				exit(1);
			}
			n = rsdata.get_num_elements();
			break;
		case CPLX_RECT:
		case CPLX_POLAR:
			if (!csdata.load_from_text_file(opts.input_file_name)) {
				fprintf(stderr, "%s:  couldn't read \"%s\".\n",
					opts.argv0, opts.input_file_name);
				exit(1);
			}
			n = csdata.get_num_elements();
			break;
		default:
			CODING_ERROR();
			break;
		}
		break;

	case FMT_HEX:
		fprintf(stderr, "hex format unimplemented.\n");
		exit(1);
		break;
	default:
		CODING_ERROR();
		break;
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Transform input data to canonical form (array of complexes in
	// rectangular format).

	switch (opts.input_cplx_format) {
	case CPLX_RE:
		csdata = tvector<complex_float>(n);
		for (i = 0; i < n; i++) {
			csdata[i].re = rsdata[i];
			csdata[i].im = 0.0;
		}
		break;
	case CPLX_IM:
		csdata = tvector<complex_float>(n);
		for (i = 0; i < n; i++) {
			csdata[i].re = 0.0;
			csdata[i].im = rsdata[i];
		}
		break;
	case CPLX_MAG:
		csdata = tvector<complex_float>(n);
		for (i = 0; i < n; i++) {
			csdata[i].re = rsdata[i];
			csdata[i].im = 0.0;
		}
		break;
	case CPLX_MAG_LOG:
		csdata = tvector<complex_float>(n);
		for (i = 0; i < n; i++) {
			csdata[i].re = pow(10.0, rsdata[i]);
			csdata[i].im = 0.0;
		}
		break;
	case CPLX_PHZ:
		csdata = tvector<complex_float>(n);
		for (i = 0; i < n; i++) {
			csdata[i].re = cos(rsdata[i]);
			csdata[i].im = sin(rsdata[i]);
		}
		break;
	case CPLX_RECT:
		// Already the way we want it.
		break;
	case CPLX_POLAR:
		for (i = 0; i < n; i++) {
			float mag = csdata[i].re;
			float phz = csdata[i].im;
			csdata[i].re = mag * cos(phz);
			csdata[i].im = mag * sin(phz);
		}
		break;
	default: CODING_ERROR();
		break;
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Transform canonical data to output format.
	switch (opts.output_cplx_format) {
	case CPLX_RE:
		rsdata = tvector<float>(n);
		for (i = 0; i < n; i++) {
			rsdata[i] = csdata[i].re;
		}
		break;
	case CPLX_IM:
		rsdata = tvector<float>(n);
		for (i = 0; i < n; i++) {
			rsdata[i] = csdata[i].im;
		}
		break;
	case CPLX_MAG:
		rsdata = tvector<float>(n);
		for (i = 0; i < n; i++) {
			rsdata[i] = csdata[i].mag();
		}
		break;
	case CPLX_MAG_LOG:
	{
		// If log is desired, we need to know what to do with log(0).
		// Making it a very negative number causes the non-zero data to
		// be squinched up at the top of a plot.  Making it not so negative
		// causes minus infinity to be greater than some of the data -- also
		// not good.  So, make one pass finding the minimum non-zero magnitude.
		// Then make a second pass, setting zero logs to minmag - 1 or so.

		float minmag = 0.01;
		float logminmag;

		rsdata = tvector<float>(n);

		// Convert to magnitude
		for (i = 0; i < n; i++)
			rsdata[i] = csdata[i].mag();

		// Find smallest non-zero magnitude
		for (i = 0; i < n; i++)
			if ((rsdata[i] != 0.0) && (rsdata[i] < minmag))
					minmag = rsdata[i];

		logminmag = log10(minmag) - 1.0;
		for (i = 0; i < n; i++) {
			if (rsdata[i] == 0.0)
				rsdata[i] = logminmag;
			else
				rsdata[i] = log10(rsdata[i]);
		}
	}
		break;
	case CPLX_PHZ:
		rsdata = tvector<float>(n);
		for (i = 0; i < n; i++) {
			rsdata[i] = csdata[i].phz();
		}
		break;
	case CPLX_RECT:
		// Already the way we want it.
		break;
	case CPLX_POLAR:
		for (i = 0; i < n; i++) {
			float mag = csdata[i].mag();
			float phz = csdata[i].phz();
			csdata[i].re = mag;
			csdata[i].im = phz;
		}
		break;
	default: CODING_ERROR();
		break;
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Write output data
	switch (opts.output_file_format) {

	case FMT_BIN:
		switch (opts.output_cplx_format) {
		case CPLX_RE:
		case CPLX_IM:
		case CPLX_MAG:
		case CPLX_MAG_LOG:
		case CPLX_PHZ:
			store_to_binary_file_or_die_nbo(rsdata,opts.output_file_name);
			break;
		case CPLX_RECT:
		case CPLX_POLAR:
			store_to_binary_file_or_die_nbo(csdata,opts.output_file_name);
			break;
		default: CODING_ERROR();
			break;
		}
		break;

	case FMT_TXT:
		switch (opts.output_cplx_format) {
		case CPLX_RE:
		case CPLX_IM:
		case CPLX_MAG:
		case CPLX_MAG_LOG:
		case CPLX_PHZ:
			rsdata.store_to_text_file(opts.output_file_name);
			break;
		case CPLX_RECT:
		case CPLX_POLAR:
			csdata.store_to_text_file(opts.output_file_name);
			break;
		default: CODING_ERROR();
			break;
		}
		break;

	case FMT_HEX:
		fprintf(stderr, "hex format unimplemented.\n");
		exit(1);
		break;
	default:
		CODING_ERROR();
		break;
	}
}

// ----------------------------------------------------------------
static void convert(
	opts_t opts)
{
	tvector<complex_double> cddata;
	tvector<complex_float>  csdata;
	tvector<double> rddata;
	tvector<float>  rsdata;
	int i, n = 0;

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Read input data
	switch (opts.input_file_format) {

	case FMT_BIN:
		switch (opts.input_cplx_format) {
		case CPLX_RE:
		case CPLX_IM:
		case CPLX_MAG:
		case CPLX_MAG_LOG:
		case CPLX_PHZ:
			if (opts.input_precision == PREC_SINGLE) {
				load_from_binary_file_or_die_nbo(rsdata, opts.input_file_name);
				n = rsdata.get_num_elements();
				rddata = tvector<double>(n);
				for (i = 0; i < n; i++)
					rddata[i] = rsdata[i];
			}
			else {
				load_from_binary_file_or_die_nbo(rddata, opts.input_file_name);
				n = rddata.get_num_elements();
			}
			break;
		case CPLX_RECT:
		case CPLX_POLAR:
			if (opts.input_precision == PREC_SINGLE) {
				load_from_binary_file_or_die_nbo(csdata,opts.input_file_name);
				n = csdata.get_num_elements();
				cddata = tvector<complex_double>(n);
				for (i = 0; i < n; i++) {
					cddata[i].re = csdata[i].re;
					cddata[i].im = csdata[i].im;
				}
			}
			else {
				load_from_binary_file_or_die_nbo(cddata, opts.input_file_name);
				n = cddata.get_num_elements();
			}
			break;
		default:
			CODING_ERROR();
			break;
		}
		break;

	case FMT_TXT:
		switch (opts.input_cplx_format) {
		case CPLX_RE:
		case CPLX_IM:
		case CPLX_MAG:
		case CPLX_MAG_LOG:
		case CPLX_PHZ:
			if (opts.input_precision == PREC_SINGLE) {
				if (!rsdata.load_from_text_file(opts.input_file_name)) {
					fprintf(stderr, "%s:  couldn't read \"%s\".\n",
						opts.argv0, opts.input_file_name);
					exit(1);
				}
				n = rsdata.get_num_elements();
				rddata = tvector<double>(n);
				for (i = 0; i < n; i++) {
					rddata[i] = rsdata[i];
				}
			}
			else {
				if (!rddata.load_from_text_file(opts.input_file_name)) {
					fprintf(stderr, "%s:  couldn't read \"%s\".\n",
						opts.argv0, opts.input_file_name);
					exit(1);
				}
				n = rddata.get_num_elements();
			}
			break;
		case CPLX_RECT:
		case CPLX_POLAR:
			if (opts.input_precision == PREC_SINGLE) {
				if (!csdata.load_from_text_file(opts.input_file_name)) {
					fprintf(stderr, "%s:  couldn't read \"%s\".\n",
						opts.argv0, opts.input_file_name);
					exit(1);
				}
				n = csdata.get_num_elements();
				cddata = tvector<complex_double>(n);
				for (i = 0; i < n; i++) {
					cddata[i].re = csdata[i].re;
					cddata[i].im = csdata[i].im;
				}
			}
			else {
				if (!cddata.load_from_text_file(opts.input_file_name)) {
					fprintf(stderr, "%s:  couldn't read \"%s\".\n",
						opts.argv0, opts.input_file_name);
					exit(1);
				}
				n = cddata.get_num_elements();
			}
			break;
		default:
			CODING_ERROR();
			break;
		}
		break;

	case FMT_HEX:
		fprintf(stderr, "hex format unimplemented.\n");
		exit(1);
		break;
	default:
		CODING_ERROR();
		break;
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Transform input data to canonical form (array of complex doubles in
	// rectangular format).

	switch (opts.input_cplx_format) {
	case CPLX_RE:
		cddata = tvector<complex_double>(n);
		for (i = 0; i < n; i++) {
			cddata[i].re = rddata[i];
			cddata[i].im = 0.0;
		}
		break;
	case CPLX_IM:
		cddata = tvector<complex_double>(n);
		for (i = 0; i < n; i++) {
			cddata[i].re = 0.0;
			cddata[i].im = rddata[i];
		}
		break;
	case CPLX_MAG:
	case CPLX_MAG_LOG:
		cddata = tvector<complex_double>(n);
		for (i = 0; i < n; i++) {
			cddata[i].re = rddata[i];
			cddata[i].im = 0.0;
		}
		break;
	case CPLX_PHZ:
		cddata = tvector<complex_double>(n);
		for (i = 0; i < n; i++) {
			cddata[i].re = cos(rddata[i]);
			cddata[i].im = sin(rddata[i]);
		}
		break;
	case CPLX_RECT:
		// Already the way we want it.
		break;
	case CPLX_POLAR:
		for (i = 0; i < n; i++) {
			float mag = cddata[i].re;
			float phz = cddata[i].im;
			cddata[i].re = mag * cos(phz);
			cddata[i].im = mag * sin(phz);
		}
		break;
	default: CODING_ERROR();
		break;
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Transform canonical data to output format.
	switch (opts.output_cplx_format) {
	case CPLX_RE:
		rddata = tvector<double>(n);
		for (i = 0; i < n; i++) {
			rddata[i] = cddata[i].re;
		}
		break;
	case CPLX_IM:
		rddata = tvector<double>(n);
		for (i = 0; i < n; i++) {
			rddata[i] = cddata[i].im;
		}
		break;
	case CPLX_MAG:
	case CPLX_MAG_LOG:
		rddata = tvector<double>(n);
		for (i = 0; i < n; i++) {
			rddata[i] = cddata[i].mag();
		}
		break;
	case CPLX_PHZ:
		rddata = tvector<double>(n);
		for (i = 0; i < n; i++) {
			rddata[i] = cddata[i].phz();
		}
		break;
	case CPLX_RECT:
		// Already the way we want it.
		break;
	case CPLX_POLAR:
		for (i = 0; i < n; i++) {
			float mag = cddata[i].mag();
			float phz = cddata[i].phz();
			cddata[i].re = mag;
			cddata[i].im = phz;
		}
		break;
	default: CODING_ERROR();
		break;
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Write output data
	switch (opts.output_file_format) {

	case FMT_BIN:
		switch (opts.output_cplx_format) {
		case CPLX_RE:
		case CPLX_IM:
		case CPLX_MAG:
		case CPLX_MAG_LOG:
		case CPLX_PHZ:
			if (opts.output_precision == PREC_SINGLE) {
				n = rddata.get_num_elements();
				rsdata = tvector<float>(n);
				for (i = 0; i < n; i++)
					rsdata[i] = rddata[i];
				store_to_binary_file_or_die_nbo(rsdata, opts.output_file_name);
			}
			else {
				store_to_binary_file_or_die_nbo(rddata, opts.output_file_name);
			}
			break;
		case CPLX_RECT:
		case CPLX_POLAR:
			if (opts.output_precision == PREC_SINGLE) {
				n = cddata.get_num_elements();
				csdata = tvector<complex_float>(n);
				for (i = 0; i < n; i++) {
					csdata[i].re = cddata[i].re;
					csdata[i].im = cddata[i].im;
				}
				store_to_binary_file_or_die_nbo(csdata, opts.output_file_name);
			}
			else {
				store_to_binary_file_or_die_nbo(cddata, opts.output_file_name);
			}
			break;
		default: CODING_ERROR();
			break;
		}
		break;

	case FMT_TXT:
		switch (opts.output_cplx_format) {
		case CPLX_RE:
		case CPLX_IM:
		case CPLX_MAG:
		case CPLX_MAG_LOG:
		case CPLX_PHZ:
			if (opts.output_precision == PREC_SINGLE) {
				n = rddata.get_num_elements();
				rsdata = tvector<float>(n);
				for (i = 0; i < n; i++)
					rsdata[i] = rddata[i];
				rsdata.store_to_text_file(opts.output_file_name);
			}
			else {
				rddata.store_to_text_file(opts.output_file_name);
			}
			break;
		case CPLX_RECT:
		case CPLX_POLAR:
			if (opts.output_precision == PREC_SINGLE) {
				n = cddata.get_num_elements();
				csdata = tvector<complex_float>(n);
				for (i = 0; i < n; i++) {
					csdata[i].re = cddata[i].re;
					csdata[i].im = cddata[i].im;
				}
				csdata.store_to_text_file(opts.output_file_name);
			}
			else {
				cddata.store_to_text_file(opts.output_file_name);
			}
			break;
		default: CODING_ERROR();
			break;
		}
		break;

	case FMT_HEX:
		fprintf(stderr, "hex format unimplemented.\n");
		exit(1);
		break;
	default:
		CODING_ERROR();
		break;
	}
}
