// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include "tvector.h"

#include "dft_spfp.h"
#include "fft_r2_spfp.h"

#include "dft_dpfp.h"
#include "fft_r2_dpfp.h"
#include "nbio.h"

#define PREC_SINGLE 30
#define PREC_DOUBLE 31

#define RADIX_2     41
#define RADIX_4     42
#define DO_DFT      43

// ----------------------------------------------------------------
struct host_fft_opts_t {
	fft_opts_t fft;
	fft_verbosity_t vopts;
	int    precision;
	int    radix;
	char * input_file_name;
	char * output_file_name;
};

// ----------------------------------------------------------------
static void usage(
	char * argv0);

#define CODING_ERROR() coding_error(__FILE__, __LINE__)
static void coding_error(
	char * file,
	int    line);

void set_defaults(
	host_fft_opts_t & ropts);

void do_single(
	host_fft_opts_t & ropts);

void do_double(
	host_fft_opts_t & ropts);

int parse_command_line(
	host_fft_opts_t & ropts,
	int     argc,
	char ** argv);

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	host_fft_opts_t opts;

	set_defaults(opts);
	if (!parse_command_line(opts, argc, argv))
		usage(argv[0]);

	switch (opts.precision) {
	case PREC_SINGLE: do_single(opts); break;
	case PREC_DOUBLE: do_double(opts); break;
	default: CODING_ERROR();           break;
	}

	return 0;
}

// ----------------------------------------------------------------
static void usage(
	char * argv0)
{
	fprintf(stderr, "Usage: %s [flags] {input file} {output file}\n", argv0);
	fprintf(stderr, "Flags:\n");
	fprintf(stderr, "  -fi:   fold input (default no).\n");
	fprintf(stderr, "  -fo:   fold output (default no).\n");
	fprintf(stderr, "  -fwd:  Forward FFT (default).\n");
	fprintf(stderr, "  -rev:  Reverse FFT.\n");
	fprintf(stderr, "  -s:    Scaled FFT (default).\n");
	fprintf(stderr, "  -ns:   No scaling.\n");
	fprintf(stderr, "  -v:    Verbosity to stderr.\n");
	fprintf(stderr, "  -vv:   Extra verbosity to stderr.\n");
	fprintf(stderr, "  -bv:   Butterfly verbosity to stderr.\n");
	fprintf(stderr, "  -sp:   I/O and FFT are single-precision .\n");
	fprintf(stderr, "  -dp:   I/O and FFT are double-precision .\n");
	fprintf(stderr, "  -r2:   Use radix-2 FFT .\n");
	fprintf(stderr, "  -dft:  Use DFT .\n");
	exit(1);
}

// ----------------------------------------------------------------
static void coding_error(
	char * file,
	int    line)
{
	fprintf(stderr, "Coding error detected in file %s, line %d\n",
		file, line);
	exit(1);
}

// ----------------------------------------------------------------
void set_defaults(
	host_fft_opts_t & ropts)
{
	ropts.fft.N                      = 0;
	ropts.fft.fold_in                = 0;
	ropts.fft.fold_out               = 0;
	ropts.fft.forward                = 1;
	ropts.fft.scale                  = 1;

	ropts.vopts.verbose              = 0;
	ropts.vopts.extra_verbose        = 0;
	ropts.vopts.butterfly_verbose    = 0;

	ropts.precision                  = PREC_SINGLE;
	ropts.radix                      = RADIX_2;

	ropts.input_file_name            = 0;
	ropts.output_file_name           = 0;
}

// ----------------------------------------------------------------
int parse_command_line(
	host_fft_opts_t & ropts,
	int     argc,
	char ** argv)
{
	int argi = 1;

	ropts.input_file_name       = 0;
	ropts.output_file_name      = 0;

	while (argi < argc) {
		if (argv[argi][0] != '-')
			break;
		if (strcmp(argv[argi], "-") == 0)
			break;

		else if (strcmp(argv[argi], "-fi") == 0)
			ropts.fft.fold_in = 1;
		else if (strcmp(argv[argi], "-fo") == 0)
			ropts.fft.fold_out = 1;
		else if (strcmp(argv[argi], "-fwd") == 0)
			ropts.fft.forward = 1;
		else if (strcmp(argv[argi], "-rev") == 0)
			ropts.fft.forward = 0;
		else if (strcmp(argv[argi], "-s") == 0)
			ropts.fft.scale = 1;
		else if (strcmp(argv[argi], "-ns") == 0)
			ropts.fft.scale = 0;

		else if (strcmp(argv[argi], "-v") == 0)
			ropts.vopts.verbose = 1;
		else if (strcmp(argv[argi], "-vv") == 0)
			ropts.vopts.verbose = ropts.vopts.extra_verbose = 1;
		else if (strcmp(argv[argi], "-bv") == 0)
			ropts.vopts.butterfly_verbose = 1;

		else if (strcmp(argv[argi], "-sp") == 0)
			ropts.precision = PREC_SINGLE;
		else if (strcmp(argv[argi], "-dp") == 0)
			ropts.precision = PREC_DOUBLE;

		else if (strcmp(argv[argi], "-r2") == 0)
			ropts.radix = RADIX_2;
		else if (strcmp(argv[argi], "-dft") == 0)
			ropts.radix = DO_DFT;

		else
			return 0;
		argi++;
	}

	if ((argc - argi) != 2)
		return 0;
	ropts.input_file_name  = argv[argi];
	ropts.output_file_name = argv[argi+1];

	return 1;
}

// ----------------------------------------------------------------
void do_single(
	host_fft_opts_t & ropts)
{
	tvector<complex_float> v;
	load_from_binary_file_or_die_nbo(v, ropts.input_file_name);
	ropts.fft.N = v.get_num_elements();

	switch (ropts.radix) {
	case RADIX_2:
		fft_r2_spfp(v.expose(), ropts.fft, ropts.vopts);
		break;
	case DO_DFT:
		dft_spfp(v.expose(), ropts.fft, ropts.vopts);
		break;
	default:
		CODING_ERROR();
		break;
	}

	store_to_binary_file_or_die_nbo(v, ropts.output_file_name);
}

// ----------------------------------------------------------------
void do_double(
	host_fft_opts_t & ropts)
{
	tvector<complex_double> v;
	load_from_binary_file_or_die_nbo(v, ropts.input_file_name);
	ropts.fft.N = v.get_num_elements();

	switch (ropts.radix) {
	case RADIX_2:
		fft_r2_dpfp(v.expose(), ropts.fft, ropts.vopts);
		break;
	case DO_DFT:
		dft_dpfp(v.expose(), ropts.fft, ropts.vopts);
		break;
	default:
		CODING_ERROR();
		break;
	}

	store_to_binary_file_or_die_nbo(v, ropts.output_file_name);
}
