// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "complex_float.h"
#include "tmatrix.h"
#include "nbio.h"

typedef struct _opts_t {
	int    do_complex;
	int    do_circular;
	int    left;
	int    right;
	int    top;
	int    bottom;
	int    num_rows;
	int    num_cols;
	char * output_file_name;
} opts_t;

static void set_defaults(opts_t * popts);
static int parse_command_line(int argc, char ** argv, opts_t * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {-r|-c} {left-right} {top-bottom} {# rows} {# cols} "
		"{output file}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	opts_t opts;
	int i, j;

	set_defaults(&opts);
	if (!parse_command_line(argc, argv, &opts))
		usage(argv[0]);

	tmatrix<float> fdata(opts.num_rows, opts.num_cols);

	if (opts.do_circular) {
		float x0 = (opts.left + opts.right)  / 2;
		float y0 = (opts.top  + opts.bottom) / 2;
		float a  = (opts.left - opts.right)  / 2;
		float b  = (opts.top  - opts.bottom) / 2;
		if (a == 0) a = 1.0;
		if (b == 0) b = 1.0;
		float a2inv = 1.0 / a / a;
		float b2inv = 1.0 / b / b;

		for (i = 0; i < opts.num_rows; i++) {
			float y = i;
			for (j = 0; j < opts.num_cols; j++) {
				float x = j;
				float dist =
					((x - x0) * (x - x0) * a2inv) +
					((y - y0) * (y - y0) * b2inv);
				if (dist <= 1.0)
					fdata[i][j] = 1.0;
				else
					fdata[i][j] = 0.0;
			}
		}
	}

	else {
		for (i = 0; i < opts.num_rows; i++) {
			for (j = 0; j < opts.num_cols; j++) {
				fdata[i][j] = 0.0;
				if (i < opts.top)
					continue;
				if (i > opts.bottom)
					continue;
				if (j < opts.left)
					continue;
				if (j > opts.right)
					continue;
				fdata[i][j] = 1.0;
			}
		}
	}

	if (opts.do_complex) {
		tmatrix<complex_float> cfdata(opts.num_rows, opts.num_cols);
		for (i = 0; i < opts.num_rows; i++)
			for (j = 0; j < opts.num_cols; j++)
				cfdata[i][j] = fdata[i][j];
		store_to_binary_file_or_die_nbo(cfdata, opts.output_file_name);
	}
	else {
		store_to_binary_file_or_die_nbo(fdata, opts.output_file_name);
	}

	return 0;
}

// ----------------------------------------------------------------
static void set_defaults(opts_t * popts)
{
	popts->do_complex       = 1;
	popts->do_circular      = 0;
	popts->left             = 0;
	popts->right            = 0;
	popts->top              = 0;
	popts->bottom           = 0;
	popts->num_rows         = 1;
	popts->num_cols         = 1;
	popts->output_file_name = 0;
}

// ----------------------------------------------------------------
static int parse_command_line(int argc, char ** argv, opts_t * popts)
{
	int argi;
	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "--help") == 0)
			return 0;
		if (argv[argi][0] != '-')
			break;

		else if (strcmp(argv[argi], "-r") == 0) {
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-c") == 0) {
			popts->do_complex = 1;
		}
		else if (strcmp(argv[argi], "-circ") == 0) {
			popts->do_circular = 1;
		}

		else {
			return 0;
		}
	}
	if ((argc - argi) != 5)
		return 0;

	if (sscanf(argv[argi], "%d-%d", &popts->left, &popts->right) != 2)
		return 0;
	if (sscanf(argv[argi+1], "%d-%d", &popts->top, &popts->bottom) != 2)
		return 0;
	if (sscanf(argv[argi+2], "%d", &popts->num_rows) != 1)
		return 0;
	if (sscanf(argv[argi+3], "%d", &popts->num_cols) != 1)
		return 0;
	popts->output_file_name = argv[argi+4];

	return 1;
}
