// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include "tvector.h"
#include "nbio.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-sp|-dp] [-std] {input file}.\n", argv0);
	fprintf(stderr, "With -std, also compute the standard deviation.\n");
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int do_double = 0;
	int do_stddev = 0;
	int argb = 1;
	char * input_file_name = 0;

	if (argc < 2)
		usage(argv[0]);
	while ((argb < argc) && (argv[argb][0] == '-')) {
		if (strcmp(argv[argb], "--help") == 0)
			usage(argv[0]);
		else if (strcmp(argv[argb], "-sp") == 0)
			do_double  = 0;
		else if (strcmp(argv[argb], "-dp") == 0)
			do_double  = 1;
		else if (strcmp(argv[argb], "-std") == 0)
			do_stddev  = 1;
		else
			usage(argv[0]);
		argb++;
	}

	if ((argc - argb) < 1)
		usage(argv[0]);

	for (int argi = argb; argi < argc; argi++) {
		input_file_name = argv[argi];

		if (do_double) {
			tvector<double> x;
			int n;
			double mean;
			load_from_binary_file_or_die_nbo(x, input_file_name);

			n = x.get_num_elements();
			mean = 0.0;
			for (int i = 0; i < n; i++)
				mean += x[i];
			mean /= n;
			printf("%s: %18.11lg", input_file_name, mean);

			if (do_stddev) {
				double stddev = 0.0;
				double curdiff;
				for (int i = 0; i < n; i++) {
					curdiff = x[i] - mean;
					stddev += curdiff * curdiff;
				}
				stddev /= n - 1; // Warning: blows up at n = 1!
				stddev = sqrt(stddev);
				printf("  %11.7lg", stddev);
			}
			printf("\n");
		}
		else {
			tvector<float> x;
			int n;
			float mean;
			load_from_binary_file_or_die_nbo(x, input_file_name);

			n = x.get_num_elements();
			mean = 0.0;
			for (int i = 0; i < n; i++)
				mean += x[i];
			mean /= n;
			printf("%s: %11.7g", input_file_name, mean);

			if (do_stddev) {
				float stddev = 0.0;
				float curdiff;
				for (int i = 0; i < n; i++) {
					curdiff = x[i] - mean;
					stddev += curdiff * curdiff;
				}
				stddev /= n - 1; // Warning: blows up at n = 1!
				stddev = sqrt(stddev);
				printf("  %11.7g", stddev);
			}
			printf("\n");
		}
	}

	return 0;
}
