// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>

static void usage(char * argv0);

static void read_ppm_header(
	char * argv0, char * ppm_input_file_name,
	char * line1, char * line2, char * line3,
	int * pnum_rows, int * pnum_cols, int * pmax_pix_val);

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	char * ppm_input_file_name;
	char * bin_output_file_name;
	FILE * infp  = 0;
	FILE * outfp = 0;
	char   line1[80];
	char   line2[80];
	char   line3[80];
	char * p1, * p2, * p3;
	int    num_rows = 0, num_cols = 0, max_pix_val = 0;
	unsigned char rgb[3];
	int    triple_counter = 0;
	float  x = 0.0;

	// Parse the command line.
	if (argc != 3)
		usage(argv[0]);
	ppm_input_file_name  = argv[1];
	bin_output_file_name = argv[2];

	// Open the input and output files.
	infp  = fopen(ppm_input_file_name, "rb");
	if (!infp) {
		fprintf(stderr,
			"%s:  Couldn't open PPM input file \"%s\" for read.\n",
			argv[0], ppm_input_file_name);
		exit(1);
	}
	outfp = fopen(bin_output_file_name, "wb");
	if (!outfp) {
		fprintf(stderr,
			"%s:  Couldn't open output binary file \"%s\" for write.\n",
			argv[0], bin_output_file_name);
		exit(1);
	}

	// Read the PPM header from the input file.
	p1 = fgets(line1, sizeof(line1), infp);
	p2 = fgets(line2, sizeof(line2), infp);
	p3 = fgets(line3, sizeof(line3), infp);
	if (!p1 || !p2 || !p3) {
		fprintf(stderr, "%s:  couldn't read PPM header from \"%s\".\n",
			argv[0], ppm_input_file_name);
		exit(1);
	}
	read_ppm_header(argv[0], ppm_input_file_name, line1, line2, line3,
		&num_rows, &num_cols, &max_pix_val);

	triple_counter = 0;
	while (fread(rgb, sizeof(rgb[0]), 3, infp) == 3) {
		x = (rgb[0] + rgb[1] + rgb[2]) / 3.0;
		*(volatile unsigned *)&x = htonl(*(volatile unsigned *)&x);
		(void)fwrite((char *)&x, sizeof(x), 1, outfp);
		triple_counter++;
	}

	if (triple_counter != num_rows * num_cols) {
		fprintf(stderr,
			"%s:  Warning:  expected %dx%d=%d RGB triples; got %d.\n",
			argv[0], num_rows, num_cols, num_rows*num_cols,
			triple_counter);
	}

	printf("%d %d\n", num_rows, num_cols);

	fclose(infp);
	fclose(outfp);

	return 0;
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s  {PPM input file} {binary output file}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void read_ppm_header(
	char * argv0, char * ppm_input_file_name,
	char * line1, char * line2, char * line3,
	int * pnum_rows, int * pnum_cols, int * pmax_pix_val)
{
	int len1, len2, len3;

	// Chomp trailing newlines
	len1 = strlen(line1);
	len2 = strlen(line2);
	len3 = strlen(line3);
	if (line1[len1] == '\n')
		line1[len1] = 0;
	if (line2[len2] == '\n')
		line2[len2] = 0;
	if (line3[len3] == '\n')
		line3[len3] = 0;

	// Validate the PPM header.  Expect something of the following form:
	//   P6
	//   60 62
	//   255
	// where "P6" is the magic number for PPM format, 60 is the number of
	// columns, 62 is the number of rows, and 255 is the maximum pixel value.

	if (strcmp(line1, "P6\n") != 0) {
		fprintf(stderr,
			"%s:  expected P6 in line 1 of PPM header of \"%s\".  Got:\n%s",
			argv0, ppm_input_file_name, line1);
		exit(1);
	}

	if (sscanf(line2, "%d %d", pnum_cols, pnum_rows) != 2) {
		fprintf(stderr,
			"%s:  couldn't parse #cols #rows in line 2 of PPM header of \"%s\".  Got:\n%s",
			argv0, ppm_input_file_name, line2);
		exit(1);
	}

	if (sscanf(line3, "%d", pmax_pix_val) != 1) {
		fprintf(stderr,
			"%s:  couldn't parse max_pix_val in line 3 of PPM header "
			"of \"%s\".  Got:\n%s",
			argv0, ppm_input_file_name, line3);
		exit(1);
	}

	if ((*pmax_pix_val < 1) || (*pmax_pix_val > 255)) {
		fprintf(stderr,
			"%s:  max_pix_val %d out of range 1-255 at line 3 of "
			"PPM header of \"%s\".\n",
			argv0, *pmax_pix_val, ppm_input_file_name);
		exit(1);
	}
}
