// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "complex_float.h"
#include "tmatrix.h"
#include "nbio.h"

typedef struct _opts_t {
	int    do_complex;
	char * ul_string;
	char * ur_string;
	char * ll_string;
	char * lr_string;
	int    num_rows;
	int    num_cols;
	char * output_file_name;
} opts_t;

static void set_defaults(opts_t * popts);
static int parse_command_line(int argc, char ** argv, opts_t * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {-r|-c} {ul} {ur} {ll} {lr} {# rows} {# cols} "
		"{output file}\n",
		argv0);
	fprintf(stderr, "Specify upper/lower left/right data values.\n");
	exit(1);
}

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	opts_t opts;
	int i, j;

	set_defaults(&opts);
	if (!parse_command_line(argc, argv, &opts))
		usage(argv[0]);

	if (opts.do_complex) {
		complex_float ul, ur, ll, lr;
		if (!ul.from_string(opts.ul_string)) usage(argv[0]);
		if (!ur.from_string(opts.ur_string)) usage(argv[0]);
		if (!ll.from_string(opts.ll_string)) usage(argv[0]);
		if (!lr.from_string(opts.lr_string)) usage(argv[0]);

		tmatrix<complex_float> fdata(opts.num_rows, opts.num_cols);

		for (i = 0; i < opts.num_rows; i++) {
			float ifracl = (float)i / (float)(opts.num_rows - 1);
			float ifracu = 1.0 - ifracl;
			for (j = 0; j < opts.num_cols; j++) {
				float jfracr = (float)j / (float)(opts.num_cols - 1);
				float jfracl = 1.0 - jfracr;
				fdata[i][j] =
					(ul * ifracu * jfracl) +
					(ll * ifracl * jfracl) +
					(ur * ifracu * jfracr) +
					(lr * ifracl * jfracr);
			}
		}
		store_to_binary_file_or_die_nbo(fdata, opts.output_file_name);
	}

	else {
		float ul, ur, ll, lr;
		if (sscanf(opts.ul_string, "%f", &ul) != 1) usage(argv[0]);
		if (sscanf(opts.ur_string, "%f", &ur) != 1) usage(argv[0]);
		if (sscanf(opts.ll_string, "%f", &ll) != 1) usage(argv[0]);
		if (sscanf(opts.lr_string, "%f", &lr) != 1) usage(argv[0]);

		tmatrix<float> fdata(opts.num_rows, opts.num_cols);

		for (i = 0; i < opts.num_rows; i++) {
			float ifracl = (float)i / (float)(opts.num_rows - 1);
			float ifracu = 1.0 - ifracl;
			for (j = 0; j < opts.num_cols; j++) {
				float jfracr = (float)j / (float)(opts.num_cols - 1);
				float jfracl = 1.0 - jfracr;
				fdata[i][j] =
					(ul * ifracu * jfracl) +
					(ll * ifracl * jfracl) +
					(ur * ifracu * jfracr) +
					(lr * ifracl * jfracr);
			}
		}
		store_to_binary_file_or_die_nbo(fdata, opts.output_file_name);
	}

	return 0;
}

// ----------------------------------------------------------------
static void set_defaults(opts_t * popts)
{
	popts->do_complex = 1;
	popts->num_rows   = 1;
	popts->num_cols   = 1;
	popts->ul_string  = 0;
	popts->ur_string  = 0;
	popts->ll_string  = 0;
	popts->lr_string  = 0;
	popts->output_file_name = 0;
}

// ----------------------------------------------------------------
static int parse_command_line(int argc, char ** argv, opts_t * popts)
{
	int argi;
	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "--help") == 0)
			return 0;
		if (argv[argi][0] != '-')
			break;

		else if (strcmp(argv[argi], "-r") == 0) {
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-c") == 0) {
			popts->do_complex = 1;
		}

		else {
			return 0;
		}
	}
	if ((argc - argi) != 7)
		return 0;
	popts->ul_string = argv[argi];
	popts->ur_string = argv[argi+1];
	popts->ll_string = argv[argi+2];
	popts->lr_string = argv[argi+3];

	if (sscanf(argv[argi+4], "%d", &popts->num_rows) != 1)
		return 0;
	if (sscanf(argv[argi+5], "%d", &popts->num_cols) != 1)
		return 0;
	popts->output_file_name = argv[argi+6];

	return 1;
}
