// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gasdev.h"
#include "tvector.h"
#include "nbio.h"

typedef struct _opts_t {
	int    do_seed;
	int    seed;
	int    do_complex;
	int    do_double;
	float  mean;
	float  stddev;
	int    num_samples;
	char * output_file_name;
} opts_t;

static void set_defaults(opts_t * popts);
static int parse_command_line(int argc, char ** argv, opts_t * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s [-r|-c] [-sp|-dp] [-s {seed}] [-m mean] [-t std dev] {num samples} {output file}\n",
		argv0);
	fprintf(stderr, "-r for real, -c for complex.\n");
	fprintf(stderr, "Seed is optional and defaults to time of day.\n");
	fprintf(stderr, "Mean is optional and defaults to 0.0.\n");
	fprintf(stderr, "Stddev is optional and defaults to 1.0.\n");
	exit(1);
}

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	opts_t opts;
	int i;

	set_defaults(&opts);
	if (!parse_command_line(argc, argv, &opts))
		usage(argv[0]);

	if (opts.do_seed)
		sran32(opts.seed);

	if (opts.do_double) {
		if (opts.do_complex) {
			tvector<complex_double> rand(opts.num_samples);
			for (i = 0; i < opts.num_samples; i++)
				rand[i] = complex_double_gasdev_ms(opts.mean, opts.stddev);
			store_to_binary_file_or_die_nbo(rand, opts.output_file_name);
		}
		else {
			tvector<double> rand(opts.num_samples);
			for (i = 0; i < opts.num_samples; i++)
				rand[i] = double_gasdev_ms(opts.mean, opts.stddev);
			store_to_binary_file_or_die_nbo(rand, opts.output_file_name);
		}
	}
	else {
		if (opts.do_complex) {
			tvector<complex_float> rand(opts.num_samples);
			for (i = 0; i < opts.num_samples; i++)
				rand[i] = complex_float_gasdev_ms(opts.mean, opts.stddev);
			store_to_binary_file_or_die_nbo(rand, opts.output_file_name);
		}
		else {
			tvector<float> rand(opts.num_samples);
			for (i = 0; i < opts.num_samples; i++)
				rand[i] = float_gasdev_ms(opts.mean, opts.stddev);
			store_to_binary_file_or_die_nbo(rand, opts.output_file_name);
		}
	}

	return 0;
}

// ----------------------------------------------------------------
static void set_defaults(opts_t * popts)
{
	popts->do_seed     = 0;
	popts->seed        = 0;
	popts->do_complex  = 1;
	popts->do_double   = 0;
	popts->mean        = 0.0;
	popts->stddev      = 1.0;
	popts->num_samples = 1;
	popts->output_file_name = 0;
}

// ----------------------------------------------------------------
static int parse_command_line(int argc, char ** argv, opts_t * popts)
{
	int argi;
	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "--help") == 0)
			return 0;
		if (argv[argi][0] != '-')
			break;

		if (strcmp(argv[argi], "-s") == 0) {
			argi++;
			if ((argc - argi) < 1)
				return 0;
			if (sscanf(argv[argi], "%i", &popts->seed) != 1)
				return 0;
			popts->do_seed = 1;
		}

		else if (strcmp(argv[argi], "-m") == 0) {
			argi++;
			if ((argc - argi) < 1)
				return 0;
			if (sscanf(argv[argi], "%f", &popts->mean) != 1)
				return 0;
		}

		else if (strcmp(argv[argi], "-t") == 0) {
			argi++;
			if ((argc - argi) < 1)
				return 0;
			if (sscanf(argv[argi], "%f", &popts->stddev) != 1)
				return 0;
		}

		else if (strcmp(argv[argi], "-r") == 0) {
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-c") == 0) {
			popts->do_complex = 1;
		}

		else if (strcmp(argv[argi], "-dp") == 0) {
			popts->do_double = 1;
		}
		else if (strcmp(argv[argi], "-sp") == 0) {
			popts->do_double = 0;
		}

		else if (strcmp(argv[argi], "-sr") == 0) {
			popts->do_double  = 0;
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-dr") == 0) {
			popts->do_double  = 1;
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-sc") == 0) {
			popts->do_double  = 0;
			popts->do_complex = 1;
		}
		else if (strcmp(argv[argi], "-dc") == 0) {
			popts->do_double  = 1;
			popts->do_complex = 1;
		}


		else {
			return 0;
		}
	}

	if ((argc - argi) != 2)
		return 0;

	if (sscanf(argv[argi], "%d", &popts->num_samples) != 1)
		return 0;
	popts->output_file_name = argv[argi+1];

	return 1;
}
