// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "complex_float.h"
#include "tmatrix.h"
#include "nbio.h"

typedef struct _opts_t {
	int    do_complex;
	int    do_circular;
	float  num_row_bumps;
	float  num_col_bumps;
	int    num_rows;
	int    num_cols;
	char * output_file_name;
} opts_t;

static void set_defaults(opts_t * popts);
static int parse_command_line(int argc, char ** argv, opts_t * popts);

// ----------------------------------------------------------------
static float sinc(float x)
{
	if (x == 0.0)
		return 1.0;
	else
		return sin(x) / x;
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-circ] [-r|-c] {# row umps} {# col bumps} {# rows} {# cols} "
		"{output file}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	opts_t opts;
	int i, j;

	set_defaults(&opts);
	if (!parse_command_line(argc, argv, &opts))
		usage(argv[0]);

	tmatrix<float> fdata(opts.num_rows, opts.num_cols);

	for (i = 0; i < opts.num_rows; i++) {
		// Scale from the range [0 .. num_rows) to the range
		// [-num_row_bumps .. num_row_bumps]
		float ifrac  = (float)i / (float)opts.num_rows;
		float iscale = -opts.num_row_bumps + 2.0 * ifrac * opts.num_row_bumps;


		for (j = 0; j < opts.num_cols; j++) {
			// Scale from the range [0 .. num_cols) to the range
			// [-num_col_bumps .. num_col_bumps]
			float jfrac  = (float)j / (float)opts.num_cols;
			float jscale = -opts.num_col_bumps + 2.0 * jfrac * opts.num_col_bumps;

			if (opts.do_circular) {
				float r = sqrt(iscale*iscale + jscale*jscale);
				fdata[i][j] = sinc(M_PI * r);
			}
			else {
				fdata[i][j] = sinc(M_PI * iscale) * sinc(M_PI * jscale);
			}
		}
	}

	if (opts.do_complex) {
		tmatrix<complex_float> cfdata(opts.num_rows, opts.num_cols);
		for (i = 0; i < opts.num_rows; i++)
			for (j = 0; j < opts.num_cols; j++)
				cfdata[i][j] = fdata[i][j];
		store_to_binary_file_or_die_nbo(cfdata, opts.output_file_name);
	}
	else {
		store_to_binary_file_or_die_nbo(fdata, opts.output_file_name);
	}

	return 0;
}

// ----------------------------------------------------------------
static void set_defaults(opts_t * popts)
{
	popts->do_complex       = 1;
	popts->do_circular      = 0;
	popts->num_row_bumps    = 1.0;
	popts->num_col_bumps    = 1.0;
	popts->num_rows         = 1;
	popts->num_cols         = 1;
	popts->output_file_name = 0;
}

// ----------------------------------------------------------------
static int parse_command_line(int argc, char ** argv, opts_t * popts)
{
	int argi;
	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "--help") == 0)
			return 0;
		if (argv[argi][0] != '-')
			break;

		else if (strcmp(argv[argi], "-r") == 0) {
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-c") == 0) {
			popts->do_complex = 1;
		}
		else if (strcmp(argv[argi], "-circ") == 0) {
			popts->do_circular = 1;
		}

		else {
			return 0;
		}
	}
	if ((argc - argi) != 5)
		return 0;

	if (sscanf(argv[argi], "%f", &popts->num_row_bumps) != 1)
		return 0;
	if (sscanf(argv[argi+1], "%f", &popts->num_col_bumps) != 1)
		return 0;
	if (sscanf(argv[argi+2], "%d", &popts->num_rows) != 1)
		return 0;
	if (sscanf(argv[argi+3], "%d", &popts->num_cols) != 1)
		return 0;
	popts->output_file_name = argv[argi+4];

	return 1;
}
