// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "tvector.h"
#include "complex_float.h"
#include "complex_double.h"
#include "comparef.h"
#include "compared.h"
#include "comparecf.h"
#include "comparecd.h"
#include "nbio.h"

void find_range_float(
	tvector<float> & v,
	float          & fmin,
	float          & fmax);

void find_range_double(
	tvector<double> & v,
	double          & fmin,
	double          & fmax);

void find_range_complex_float(
	tvector<complex_float> & v,
	float                  & fmin,
	float                  & fmax);

void find_range_complex_double(
	tvector<complex_double> & v,
	double                  & fmin,
	double                  & fmax);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {-rs|-rd|-cs|-cd} {input file}\n",
		argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	char * input_file_name;
	int    do_complex = 0;
	int    do_double = 0;
	int    argb = 1;

	if (argc < 2)
		usage(argv[0]);
	while ((argb < argc) && (argv[argb][0] == '-')) {
		if (strcmp(argv[argb], "--help") == 0)
			usage(argv[0]);

		else if (strcmp(argv[argb], "-rs") == 0) {
			do_complex = 0;
			do_double  = 0;
		}
		else if (strcmp(argv[argb], "-rd") == 0) {
			do_complex = 0;
			do_double  = 1;
		}
		else if (strcmp(argv[argb], "-cs") == 0) {
			do_complex = 1;
			do_double  = 0;
		}
		else if (strcmp(argv[argb], "-cd") == 0) {
			do_complex = 1;
			do_double  = 1;
		}

		else
			usage(argv[0]);
		argb++;
	}

	if ((argc - argb) < 1)
		usage(argv[0]);

	for (int argi = argb; argi < argc; argi++) {
		input_file_name = argv[argi];

		if (do_complex) {
			if (do_double) {
				tvector<complex_double> v;
				double  fmin, fmax;
				load_from_binary_file_or_die_nbo(v, input_file_name);
				find_range_complex_double(v, fmin, fmax);
				printf("%s: min %11.7lg max %11.7lg\n",
					input_file_name, fmin, fmax);
			}
			else {
				tvector<complex_float> v;
				float fmin, fmax;
				load_from_binary_file_or_die_nbo(v, input_file_name);
				find_range_complex_float(v, fmin, fmax);
				printf("%s: min %11.7g max %11.7g\n",
					input_file_name, fmin, fmax);
			}
		}

		else {
			if (do_double) {
				tvector<double> v;
				double fmin, fmax;
				load_from_binary_file_or_die_nbo(v, input_file_name);
				find_range_double(v, fmin, fmax);
				printf("%s: min %11.7lg max %11.7lg\n",
					input_file_name, fmin, fmax);
			}
			else {
				tvector<float> v;
				float fmin, fmax;
				load_from_binary_file_or_die_nbo(v, input_file_name);
				find_range_float(v, fmin, fmax);
				printf("%s: min %11.7g max %11.7g\n",
					input_file_name, fmin, fmax);
			}
		}
	}

	return 0;
}

// ----------------------------------------------------------------
void find_range_float(
	tvector<float> & v,
	float          & fmin,
	float          & fmax)
{
	int n = v.get_num_elements();

	fmin = v[0];
	fmax = v[0];
	for (int i = 0; i < n; i++) {
		float fcur = v[i];
		if (fcur < fmin)
			fmin = fcur;
		if (fcur > fmax)
			fmax = fcur;
	}
}

// ----------------------------------------------------------------
void find_range_double(
	tvector<double> & v,
	double          & fmin,
	double          & fmax)
{
	int n = v.get_num_elements();

	fmin = v[0];
	fmax = v[0];

	for (int i = 0; i < n; i++) {
		double fcur = v[i];
		if (fcur < fmin)
			fmin = fcur;
		if (fcur > fmax)
			fmax = fcur;
	}
}

// ----------------------------------------------------------------
void find_range_complex_float(
	tvector<complex_float> & v,
	float                  & fmin,
	float                  & fmax)
{
	int n = v.get_num_elements();

	fmin = v[0].mag();
	fmax = fmin;
	for (int i = 0; i < n; i++) {
		float cur = v[i].mag();
		if (cur < fmin)
			fmin = cur;
		if (cur > fmax)
			fmax = cur;
	}
}

// ----------------------------------------------------------------
void find_range_complex_double(
	tvector<complex_double> & v,
	double                  & fmin,
	double                  & fmax)
{
	int n = v.get_num_elements();

	fmin = v[0].mag();
	fmax = fmin;
	for (int i = 0; i < n; i++) {
		double cur = v[i].mag();
		if (cur < fmin)
			fmin = cur;
		if (cur > fmax)
			fmax = cur;
	}
}
