// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

// Copyright 2004 John Kerl
// kerl at mathpost dot asu dot edu

#ifndef TMATRIX_H
#define TMATRIX_H

#include "tvector.h"
#include "linescan.h"

template <class element_type>
class tmatrix
{
public:
// ================================================================
// Constructors

// ----------------------------------------------------------------
// Default ctor.
tmatrix(void)
{
	this->nullify();
}

// ----------------------------------------------------------------
// Ctor with specified dimensions.
tmatrix(int arg_num_rows, int arg_num_cols)
{
	this->allocate(arg_num_rows, arg_num_cols);
}

// ----------------------------------------------------------------
// Ctor with specified dimensions and initial value.
tmatrix(element_type e, int arg_num_rows, int arg_num_cols)
{
	this->allocate(arg_num_rows, arg_num_cols);
	for (int i = 0; i < arg_num_rows; i++)
		for (int j = 0; j < arg_num_cols; j++)
			this->rows[i][j] = e;
}

// ----------------------------------------------------------------
// Copy ctor
tmatrix(const tmatrix<element_type> & that)
{
	this->num_rows = that.num_rows;
	this->num_cols = that.num_cols;
	this->rows = new tvector<element_type>[that.num_rows];
	for (int i = 0; i < that.num_rows; i++)
		this->rows[i] = that.rows[i];
}

// ----------------------------------------------------------------
// Dtor
~tmatrix(void)
{
	this->mfree();
}

// ================================================================
// Operators

// ----------------------------------------------------------------
// Matrix assignment from matrix
tmatrix<element_type> &
operator=(tmatrix<element_type> that)
{
	int i;

	if ((this->num_rows == that.num_rows) &&
		(this->num_cols == that.num_cols))
	{
		// Storage is fine as is.
	}
	else {
		this->mfree();
		this->num_rows = that.num_rows;
		this->num_cols = that.num_cols;
		this->rows = new tvector<element_type>[that.num_rows];
		for (i = 0; i < that.num_rows; i++)
			this->rows[i] = tvector<element_type>(that.num_cols);
	}

	for (i = 0; i < that.num_rows; i++)
		this->rows[i] = that.rows[i];
	return *this;
}

// ----------------------------------------------------------------
// Matrix assignment from scalar
tmatrix<element_type> &
operator=(const element_type scalar)
{
	if (this->rows) {
		for (int i = 0; i < this->num_rows; i++)
			for (int j = 0; j < this->num_cols; j++)
				this->rows[i][j] = scalar;
	}
	else {
		this->num_rows = 1;
		this->num_cols = 1;
		this->rows = new tvector<element_type>[1];
		this->rows[0] = tvector<element_type>(1);
		this->rows[0][0] = scalar;
	}
	return *this;
}

// ----------------------------------------------------------------
int operator==(tmatrix<element_type> that)
{
	this->check_dims(that, "operator==");
	for (int i = 0; i < this->num_rows; i++)
		if (this->rows[i] != that.rows[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int operator==(element_type e)
{
	for (int i = 0; i < this->num_rows; i++)
		if (this->rows[i] != e)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int operator!=(tmatrix<element_type> that)
{
	return !(*this == that);
}

// ----------------------------------------------------------------
int operator!=(element_type e)
{
	return !(*this == e);
}

// ----------------------------------------------------------------
// Array operator, with bounds checking.
tvector<element_type> & operator[](int row_index)
{
	if (this->num_rows == 0) {
		std::cerr << "tmatrix:  attempt to index an empty matrix.\n";
		exit(1);
	}
	if ((row_index < 0) || (row_index >= this->num_rows)) {
		std::cerr
			<< "tmatrix array operator: row index "
			<< row_index
			<< " out of bounds "
			<< 0
			<< ":"
			<< (this->num_rows - 1)
			<< std::endl;
		exit(1);
	}
	return this->rows[row_index];
}

// ----------------------------------------------------------------
// Matrix plus matrix.
tmatrix<element_type> operator+(
	tmatrix<element_type> that)
{
	this->check_dims(that, "operator+");
	tmatrix<element_type> rv(this->num_rows, this->num_cols);
	for (int i = 0; i < this->num_rows; i++)
		for (int j = 0; j < this->num_cols; j++)
			rv[i][j] = this->rows[i][j] + that.rows[i][j];
	return rv;
}

// ----------------------------------------------------------------
// Matrix minus matrix.
tmatrix<element_type> operator-(
	tmatrix<element_type> that)
{
	if ((this->num_rows != that.num_rows)
	||  (this->num_cols != that.num_cols)) {
		std::cerr
			<< "tmatrix operator-():  Incompatibly sized "
			<< "arguments ("
			<< this->num_rows << "x" << this->num_cols << ", "
			<< that.num_rows << "x" << that.num_cols << ")."
			<< std::endl;
		exit(1);
	}

	tmatrix<element_type> rv(this->num_rows, this->num_cols);
	for (int i = 0; i < this->num_rows; i++)
		for (int j = 0; j < this->num_cols; j++)
			rv[i][j] = this->rows[i][j] - that.rows[i][j];
	return rv;
}

// ----------------------------------------------------------------
// Unary matrix negation.
tmatrix<element_type> operator-(
	void)
{
	tmatrix<element_type> rv(this->num_rows, this->num_cols);
	for (int i = 0; i < this->num_rows; i++)
		for (int j = 0; j < this->num_cols; j++)
			rv[i][j] = -this->rows[i][j];
	return rv;
}

// ----------------------------------------------------------------
// Matrix plus scalar.
tmatrix<element_type> operator+(
	element_type e)
{
	this->check_square("tmatrix plus scalar");
	tmatrix<element_type> rv(*this);
	for (int i = 0; i < this->num_rows; i++)
		rv[i][i] += e;
	return rv;
}

// ----------------------------------------------------------------
// Matrix minus scalar.
tmatrix<element_type> operator-(
	element_type e)
{
	this->check_square("tmatrix minus scalar");
	tmatrix<element_type> rv(*this);
	for (int i = 0; i < this->num_rows; i++)
		rv[i][i] -= e;
	return rv;
}

// ----------------------------------------------------------------
// Matrix times vector.
tvector<element_type> operator*(
	tvector<element_type> v)
{
	int i, j;
	int v_num_elements = v.get_num_elements();

	if (this->num_cols != v_num_elements) {
		std::cerr
			<< "tmatrix operator*(): Incompatibly dimensioned "
			<< "operands ("
			<< this->num_rows << "x" << this->num_cols << ","
			<< v_num_elements << ")." << std::endl;
		exit(1);
	}

	tvector<element_type> rv(this->num_rows);
	element_type t00 = this->rows[0][0];
	element_type zero = t00 - t00;
	for (i = 0; i < this->num_rows; i++)
		rv[i] = zero;

	for (i = 0; i < this->num_rows; i++)
		for (j = 0; j < this->num_cols; j++)
			rv[i] += this->rows[i][j] * v[j];

	return rv;
}

// ----------------------------------------------------------------
// Matrix times matrix.
tmatrix<element_type> operator*(
	tmatrix<element_type> that)
{
	int i, j, k;

	if (this->num_cols != that.num_rows) {
		std::cerr
			<< "tmatrix operator*(): Incompatibly dimensioned "
			<< "operands ("
			<< this->num_rows << "x" << this->num_cols << ","
			<< that.num_rows << "x" << that.num_cols << ")."
			<< std::endl;
		exit(1);
	}

	tmatrix<element_type> rv(this->num_rows, that.num_cols);
	element_type t00 = this->rows[0][0];
	element_type zero = t00 - t00;
	for (i = 0; i < this->num_rows; i++)
		for (j = 0; j < that.num_cols; j++)
			//rv[i][j] = element_type(0);
			rv[i][j] = zero;

	for (i = 0; i < this->num_rows; i++)
		for (j = 0; j < that.num_cols; j++)
			for (k = 0; k < this->num_cols; k++)
				rv[i][j] += this->rows[i][k] * that.rows[k][j];

	return rv;
}

// ----------------------------------------------------------------
// Matrix times scalar.
tmatrix<element_type> operator*(
	element_type e)
{
	tmatrix<element_type> rv(*this);
	for (int i = 0; i < this->num_rows; i++)
		for (int j = 0; j < this->num_cols; j++)
			rv[i][j] *= e;
	return rv;
}

// ----------------------------------------------------------------
tmatrix<element_type> & operator+=(
	tmatrix<element_type> that)
{
	*this = *this + that;
	return *this;
}

tmatrix<element_type> & operator-=(
	tmatrix<element_type> that)
{
	*this = *this - that;
	return *this;
}

tmatrix<element_type> & operator+=(
	element_type e)
{
	*this = *this + e;
	return *this;
}

tmatrix<element_type> & operator-=(
	element_type e)
{
	*this = *this - e;
	return *this;
}

tvector<element_type> & operator*=(
	tvector<element_type> v)
{
	*this = *this * v;
	return *this;
}

tmatrix<element_type> & operator*=(
	tmatrix<element_type> that)
{
	*this = *this * that;
	return *this;
}

tmatrix<element_type> & operator*=(
	element_type scalar)
{
	*this = *this * scalar;
	return *this;
}

// ----------------------------------------------------------------
// Outer product of two vectors.
static tmatrix<element_type> outer(
	tvector<element_type> & u,
	tvector<element_type> & v)
{
	int m = u.get_num_elements();
	int n = v.get_num_elements();
	tmatrix<element_type> rv(m, n);
	for (int i = 0; i < m; i++)
		for (int j = 0; j < n; j++)
			rv[i][j] = u[i] * v[j];
	return rv;
}

// ----------------------------------------------------------------
private:
void check_column_index(int j)
{
	if (this->num_cols == 0) {
		std::cerr << "tmatrix:  attempt to index an empty matrix.\n";
		exit(1);
	}
	if ((j < 0) || (j >= this->num_cols)) {
		std::cerr
			<< "tmatrix: column index "
			<< j
			<< " out of bounds "
			<< 0
			<< ":"
			<< (this->num_cols - 1)
			<< std::endl;
		exit(1);
	}
}
public:

// Copy of column (*not* a reference to it).
tvector<element_type> get_column(int j)
{
	this->check_column_index(j);
	tvector<element_type> rv (this->num_rows);
	for (int i = 0; i < this->num_rows; i++)
		rv[i] = this->rows[i][j];
	return rv;
}

void put_column(int j, tvector<element_type> & v)
{
	if (v.get_num_elements() != this->num_rows) {
		std::cerr
			<< "tmatrix::put_column(): # rows "
			<< this->num_rows
			<< " != vector length "
			<< v.get_num_elements()
			<< ".\n";
		exit(1);
	}
	for (int i = 0; i < this->num_rows; i++)
		this->rows[i][j] = v[i];
}

// ----------------------------------------------------------------
tmatrix<element_type> transpose(void)
{
	tmatrix<element_type> rv(this->num_cols, this->num_rows);
	for (int i = 0; i < this->num_rows; i++)
		for (int j = 0; j < this->num_cols; j++)
			rv.rows[j][i] = this->rows[i][j];
	return rv;
}


// ----------------------------------------------------------------
tmatrix<element_type> compute_AtA(
	void)
{
	int nr = this->get_num_rows();
	int nc = this->get_num_cols();
	element_type zero = this->rows[0][0] - this->rows[0][0];
	tmatrix<element_type> AtA(zero, nc, nc);
	element_type ** iptrs  = this->expose();
	element_type ** optrs = AtA.expose();
	for (int i = 0; i < nc; i++)
		for (int j = 0; j < nc; j++)
			for (int k = 0; k < nr; k++)
				optrs[i][j] += iptrs[k][i] * iptrs[k][j];
	delete [] iptrs;
	delete [] optrs;
	return AtA;
}

// ----------------------------------------------------------------
void transpose_in_place(void)
{
	this->check_square("transpose_in_place");
	element_type temp;
	for (int i = 1; i < this->num_rows; i++) {
		for (int j = 0; j < i; j++) {
			temp             = this->rows[i][j];
			this->rows[i][j] = this->rows[j][i];
			this->rows[j][i] = temp;
		}
	}
}

// ----------------------------------------------------------------
// Instantiable only for complex types.
tmatrix<element_type> conjugate_transpose(void)
{
	tmatrix<element_type> rv(this->num_cols, this->num_rows);
	for (int i = 0; i < this->num_rows; i++) {
		for (int j = 0; j < this->num_cols; j++) {
			rv.rows[j][i].re =  this->rows[i][j].re;
			rv.rows[j][i].im = -this->rows[i][j].im;
		}
	}
	return rv;
}

// ----------------------------------------------------------------
// Instantiable only for complex types.
void conjugate_transpose_in_place(void)
{
	this->check_square("conjugate_transpose_in_place");
	element_type temp;
	for (int i = 1; i < this->num_rows; i++) {
		for (int j = 0; j < i; j++) {
			temp             = this->rows[i][j];
			this->rows[i][j] = this->rows[j][i];
			this->rows[j][i] = temp;
		}
	}
	for (int i = 0; i < this->num_rows; i++)
		for (int j = 0; j < this->num_cols; j++)
			this->rows[i][j].im = -this->rows[i][j].im;
}

// ----------------------------------------------------------------
int is_square(void)
{
	if (this->num_rows == this->num_cols)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
int check_square(char * msg)
{
	if (!this->is_square()) {
		std::cerr << msg << ":  non-square input.\n";
		exit(1);
	}
}

// ----------------------------------------------------------------
int get_num_rows(void)
{
	return this->num_rows;
}

// ----------------------------------------------------------------
int get_num_cols(void)
{
	return this->num_cols;
}


// ----------------------------------------------------------------
// The caller must delete[] the return value.

element_type ** expose(void)
{
	element_type ** ptrs = new element_type * [this->num_rows];
	for (int i = 0; i < this->num_rows; i++)
		ptrs[i] = this->rows[i].expose();
	return ptrs;
}

// ================================================================
//friend std::ostream & operator<< <>(
//	std::ostream & os,
//	const tmatrix<element_type> & m);

//friend std::istream & operator>> <>(
//	std::istream & is,
//	tmatrix<element_type> & m);

// ----------------------------------------------------------------
int load_from_text_file(char * file_name)
{
	if ((strcmp(file_name, "-") == 0) || (strcmp(file_name, "@") == 0)) {
		std::cin >> *this;
		if (std::cin.fail())
			return 0;
		else
			return 1;
	}

	std::ifstream ifs;
	ifs.open(file_name, std::ifstream::in);

	if (ifs.fail()) {
		std::cerr << "tmatrix::load_from_text_file:  couldn't open \""
			<< file_name << "\"\n";
		return 0;
	}

	ifs >> *this;

	if (ifs.fail()) {
		std::cerr << "tmatrix::load_from_text_file:  scan failure reading \""
			<< file_name << "\"\n";
		return 0;
	}
	return 1;
}

// ----------------------------------------------------------------
void load_from_text_file_or_die(char * file_name)
{
	if (!load_from_text_file(file_name)) {
		std::cerr << "Couldn't load from \"" << file_name << "\".\n";
		exit(1);
	}
}

// ----------------------------------------------------------------
int load_from_binary_file(
	char * input_file_name,
	int arg_num_rows,
	int arg_num_cols)
{
	if ((this->num_rows != arg_num_rows)
		|| (this->num_cols != arg_num_cols))
	{
		this->mfree();
		this->allocate(arg_num_rows, arg_num_cols);
	}

	int file_num_bytes;
	struct stat stat_buf;

	if (stat(input_file_name, &stat_buf) != 0) {
		perror("stat");
		fprintf(stderr,
			"tmatrix::load_from_binary_file: couldn't stat \"%s\".\n",
				input_file_name);
		return 0;
	}
	file_num_bytes =  stat_buf.st_size;

	int element_size = sizeof(element_type);
	if ((file_num_bytes % element_size) != 0) {
		fprintf(stderr, "tmatrix::load_from_binary_file: element size %d does "
			"not divide file size %d.\n",
			element_size, file_num_bytes);
		exit(1);
	}

	int expected_size = this->num_rows * this->num_cols * element_size;
	// You may wish to change "!=" to "<".
#if 0
	if (file_num_bytes != expected_size)
#else
	if (file_num_bytes < expected_size)
#endif
	{
		fprintf(stderr,
			"tmatrix::load_from_binary_file: expected file size %d; got %d.\n",
			expected_size, file_num_bytes);
		exit(1);
	}

	FILE * fp = fopen(input_file_name, "r");
	if (fp == 0) {
		perror("fopen");
		fprintf(stderr,
			"tmatrix::load_from_binary_file: couldn't open \"%s\".\n",
				input_file_name);
		return 0;
	}
	int rv = 1;
	for (int i = 0; i < num_rows; i++) {
		if (!this->rows[i].load_from_binary_stream(fp)) {
			rv = 0;
			break;
		}
	}
	fclose(fp);
	return rv;
}

// ----------------------------------------------------------------
// The matrix must already be sized as desired.
int load_from_binary_file(
	char * input_file_name)
{
	return this->load_from_binary_file(input_file_name,
		this->num_rows, this->num_cols);
}

// ----------------------------------------------------------------
// The matrix must already be sized as desired.
void load_from_binary_file_or_die(
	char * input_file_name)
{
	if (!this->load_from_binary_file(input_file_name)) {
		fprintf(stderr, "Couldn't load from binary file \"%s\".\n",
			input_file_name);
		exit(1);
	}
}

// ----------------------------------------------------------------
int store_to_binary_file(
	char * output_file_name)
{
	FILE * fp = fopen(output_file_name, "w");
	if (fp == 0)
		return 0;

	int rv = 1;
	for (int i = 0; i < this->num_rows; i++) {
		if (!this->rows[i].store_to_binary_stream(fp)) {
			rv = 0;
			break;
		}
	}
	fclose(fp);
	return rv;
}

// ----------------------------------------------------------------
void store_to_binary_file_or_die(
	char * output_file_name)
{
	if (!this->store_to_binary_file(output_file_name)) {
		fprintf(stderr, "Couldn't write to \"%s\".\n",
			output_file_name);
		exit(1);
	}
}

// ================================================================
private:
tvector<element_type> * rows;
int           num_rows;
int           num_cols;

// ----------------------------------------------------------------
void allocate(int arg_num_rows, int arg_num_cols)
{
	if ((arg_num_rows <= 0) || (arg_num_cols <= 0)) {
		std::cerr
			<< "tmatrix:  Matrix dimensions must be >= 0; got "
			<< arg_num_rows << " x " << arg_num_cols
			<< ".  Exiting.\n";
		exit(1);
	}

	this->num_rows = arg_num_rows;
	this->num_cols = arg_num_cols;
	this->rows = new tvector<element_type>[arg_num_rows];
	for (int i = 0; i < arg_num_rows; i++)
		this->rows[i] = tvector<element_type>(arg_num_cols);
}

// ----------------------------------------------------------------
void mfree(void)
{
	if (this->rows != 0)
		delete [] this->rows;
	this->nullify();
}

// ----------------------------------------------------------------
void nullify(void)
{
	this->rows = 0;
	this->num_rows = 0;
	this->num_cols = 0;
}

// ----------------------------------------------------------------
void check_dims(tmatrix<element_type> that, char * msg)
{
	if ((this->num_rows != that.num_rows)
	||  (this->num_cols != that.num_cols)) {
		std::cerr
			<< "tmatrix "
			<< msg
			<< ":  Incompatibly sized arguments ("
			<< this->num_rows << "x" << this->num_cols << ", "
			<< that.num_rows << "x" << that.num_cols << ")." << std::endl;
		exit(1);
	}
}
};

// ================================================================
template <class element_type>
static std::ostream & operator<< (
	std::ostream & os,
	const tmatrix<element_type> & m)
{
	for (int i = 0; i < m.num_rows; i++) {
		m.rows[i].row_output(os);
		os << std::endl;
	}
	return os;
}

// ================================================================
template <class element_type>
static std::istream & operator>> (
	std::istream & is,
	tmatrix<element_type> & m)
{
	char line[2048];

	const int init_num_rows = 10;
	const int more_num_rows = 10;
	int alloc_num_rows = init_num_rows;
	m.mfree();
	m.rows = new tvector<element_type>[init_num_rows];

	while (1) {
		if (is.eof())
			break;

		if (m.num_rows >= alloc_num_rows) {
			alloc_num_rows += more_num_rows;
			tvector<element_type> * ptemp =
				new tvector<element_type>[alloc_num_rows];
			for (int i = 0; i < m.num_rows; i++)
				ptemp[i] = m.rows[i];
			delete [] m.rows;
			m.rows = ptemp;
		}

		// Read a line of text.
		is.getline(line, sizeof(line));
		if (is.eof()) {
			is.clear();
			break;
		}
		else if (is.fail()) {
			break;
		}

		// Allow comments.
		char * phash = strstr(line, "#");
		if (phash)
			*phash = 0;

		// Allow multiple matrices in the same stream, delimited by
		// carriage returns.
		if (is_whitespace_line(line)) {
			if (m.num_rows == 0)
				continue;
			else
				break;
		}

		// Use tvector istringstream >> to scan the line of text as a
		// row vector.
		std::istringstream iss(line, std::ios_base::in);
		iss >> m.rows[m.num_rows];
		if (iss.fail()) {
			is.setstate(std::ios::failbit);
			std::cerr << "tmatrix scan failure: \""
				<< line << "\"\n";
			return is;
		}

		m.num_rows++;
	}

	// xxx temp
	if (is.peek() == EOF)
		is.setstate(std::ios::eofbit);

	if (m.num_rows == 0) {
		is.setstate(std::ios::badbit);
		std::cerr << "tmatrix >>:  zero-row matrix.\n";
		return is;
	}

	int min_cols = 0x7fffffff;
	int max_cols = 0;
	for (int i = 0; i < m.num_rows; i++) {
		int cur_cols = m.rows[i].get_num_elements();
		if (cur_cols > max_cols)
			max_cols = cur_cols;
		if (cur_cols < min_cols)
			min_cols = cur_cols;
	}
	if (min_cols != max_cols) {
		is.setstate(std::ios::badbit);
		std::cerr
			<< "tmatrix >>:  ragged input.  # rows = "
			<< m.num_rows
			<< " min # cols = "
			<< min_cols
			<< " max # cols = "
			<< max_cols
			<< ".\n";
		return is;
	}
	m.num_cols = max_cols;

	return is;
}

#endif // TMATRIX_H
