// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

// Copyright 2004 John Kerl
// kerl at mathpost dot asu dot edu

#ifndef TVECTOR_H
#define TVECTOR_H

#include <iostream>
#include <sstream>
#include <fstream>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h> // For ntohl() et al.

#include "complex_float.h"
#include "complex_double.h"
#include "linescan.h"

// ================================================================
template <class element_type>
class tvector
{
public:
// ================================================================
// Constructors

// ----------------------------------------------------------------
// Default ctor.
tvector(void)
{
	this->num_elements = 0;
	this->elements = 0;
}

// ----------------------------------------------------------------
// Ctor with specified size.
tvector(int init_num_elements)
{
	if (init_num_elements <= 0) {
		std::cerr
			<< "tvector::tvector():  Vector size must be > 0; got "
			<< init_num_elements
			<< ".  Exiting."
			<< std::endl;
		exit(1);
	}

	this->num_elements = init_num_elements;
	this->elements = new element_type[init_num_elements];
}

// ----------------------------------------------------------------
// Ctor from specified size with specified initial value.
tvector(element_type e, int init_num_elements)
{
	if (init_num_elements <= 0) {
		std::cerr
			<< "tvector::tvector():  Vector size must be > 0; got "
			<< init_num_elements
			<< ".  Exiting."
			<< std::endl;
		exit(1);
	}

	this->num_elements = init_num_elements;
	this->elements = new element_type[init_num_elements];
	for (int i = 0; i < this->num_elements; i++)
		this->elements[i] = e;
}

// ----------------------------------------------------------------
// Ctor from standard C++ array.
tvector(element_type * array, int init_num_elements)
{
	if (init_num_elements <= 0) {
		std::cerr
			<< "tvector::tvector():  Vector size must be > 0; got "
			<< init_num_elements
			<< ".  Exiting."
			<< std::endl;
		exit(1);
	}

	this->num_elements = init_num_elements;
	this->elements = new element_type[init_num_elements];
	for (int i = 0; i < init_num_elements; i++)
		this->elements[i] = array[i];
}

// ----------------------------------------------------------------
// Copy ctor
tvector(const tvector<element_type> & that)
{
	this->num_elements = that.num_elements;
	this->elements = new element_type[that.num_elements];
	for (int i = 0; i < that.num_elements; i++)
		this->elements[i] = that.elements[i];
}

// ----------------------------------------------------------------
// Dtor
~tvector(void)
{
	if (this->elements != 0) {
		delete [] this->elements;
		this->elements = 0;
	}
	this->num_elements = 0;
}

// ================================================================
// Operators

// ----------------------------------------------------------------
// Vector assignment from vector
tvector<element_type> &
operator=(tvector<element_type> that)
{
	this->num_elements = that.num_elements;
	if (this->elements == 0) {
		this->elements = new element_type[that.num_elements];
	}
	else if (this->num_elements < that.num_elements) {
		delete [] this->elements;
		this->elements = new element_type[that.num_elements];
	}
	for (int i = 0; i < that.num_elements; i++)
		this->elements[i] = that.elements[i];
	return *this;
}

// ----------------------------------------------------------------
// Vector assignment from scalar
tvector<element_type> &
operator=(const element_type scalar)
{
	if (this->elements) {
		for (int i = 0; i < this->num_elements; i++)
			this->elements[i] = scalar;
	}
	else {
		this->num_elements = 1;
		this->elements = new element_type[1];
		this->elements[0] = scalar;
	}
	return *this;
}

// ----------------------------------------------------------------
// Array operator with bounds checking.
element_type & operator[](int index)
{
	if (this->num_elements == 0) {
		std::cerr << "tvector:  attempt to index an empty vector.\n";
		exit(1);
	}
	if ((index < 0) || (index >= this->num_elements)) {
		std::cerr
			<< "tvector array operator: index "
			<< index
			<< " out of bounds "
			<< 0
			<< ":"
			<< (this->num_elements - 1)
			<< std::endl;
		exit(1);
	}
	return this->elements[index];
}

// ----------------------------------------------------------------
// Vector plus vector.
tvector<element_type> operator+(
	tvector<element_type> that)
{
	this->check_equal_lengths(that);
	tvector<element_type> rv(this->num_elements);
	for (int i = 0; i < this->num_elements; i++)
		rv[i] = this->elements[i] + that.elements[i];
	return rv;
}

// ----------------------------------------------------------------
// Vector minus vector.
tvector<element_type> operator-(
	tvector<element_type> that)
{
	this->check_equal_lengths(that);
	tvector<element_type> rv(this->num_elements);
	for (int i = 0; i < this->num_elements; i++)
		rv[i] = this->elements[i] - that.elements[i];
	return rv;
}

// ----------------------------------------------------------------
// Vector times scalar.
tvector<element_type> operator*(
	element_type s)
{
	tvector<element_type> rv(this->num_elements);
	for (int i = 0; i < this->num_elements; i++)
		rv[i] = this->elements[i] * s;
	return rv;
}

// ----------------------------------------------------------------
// This is componentwise multiplication (u * v).
// Use dot() (e.g. u.dot(v)) for inner product, or tmatrix's outer()
// (e.g. tmatrix::outer(u, v)) for outer product.
tvector<element_type> operator*(
	tvector<element_type> that)
{
	this->check_equal_lengths(that);
	tvector<element_type> rv(this->num_elements);
	for (int i = 0; i < this->num_elements; i++)
		rv[i] = this->elements[i] * that.elements[i];
	return rv;
}

// ----------------------------------------------------------------
// Vector dot vector.
element_type dot(
	tvector<element_type> that)
{
	this->check_equal_lengths(that);
	element_type rv = this->elements[0] - this->elements[0];
	for (int i = 0; i < this->num_elements; i++)
		rv += this->elements[i] * that.elements[i];
	return rv;
}

// ----------------------------------------------------------------
// Componentwise division.
tvector<element_type> operator/(
	tvector<element_type> that)
{
	this->check_equal_lengths(that);
	tvector<element_type> rv(this->num_elements);
	for (int i = 0; i < this->num_elements; i++)
		rv[i] = this->elements[i] / that.elements[i];
	return rv;
}

// ----------------------------------------------------------------
tvector<element_type> operator+=(
	tvector<element_type> that)
{
	*this = *this + that;
	return *this;
}
tvector<element_type> operator-=(
	tvector<element_type> that)
{
	*this = *this - that;
	return *this;
}
tvector<element_type> operator*=(
	element_type e)
{
	for (int i = 0; i < this->num_elements; i++)
		this->elements[i] *= e;
	return *this;
}
tvector<element_type> operator*=(
	tvector<element_type> that)
{
	this->check_equal_lengths(that);
	for (int i = 0; i < this->num_elements; i++)
		this->elements[i] *= that.elements[i];
	return *this;
}

// ----------------------------------------------------------------
int operator==(
	tvector<element_type> & that)
{
	if (this->num_elements != that.num_elements)
		return 0;
	for (int i = 0; i < this->num_elements; i++)
		if (this->elements[i] != that.elements[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int operator!=(
	tvector<element_type> & that)
{
	return !(*this == that);
}

// ----------------------------------------------------------------
int operator==(
	element_type e)
{
	for (int i = 0; i < this->num_elements; i++)
		if (this->elements[i] != e)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int operator!=(
	element_type e)
{
	return !(*this == e);
}

// ----------------------------------------------------------------
void check_equal_lengths(tvector<element_type> & that)
{
	if (this->num_elements != that.num_elements) {
		std::cerr
			<< "tvector operator+():  Incompatibly sized arguments ("
			<< this->num_elements << ", "
			<< that.num_elements << ")." << std::endl;
		exit(1);
	}
}

// ----------------------------------------------------------------
int get_num_elements(void)
{
	return this->num_elements;
}

// ----------------------------------------------------------------
element_type * expose(void)
{
	return this->elements;
}

// ----------------------------------------------------------------
void ptrswap(tvector<element_type> & that)
{
	if (this->num_elements != that.num_elements)
		(void)check_equal_lengths(that);
	element_type * temp = this->elements;
	this->elements = that.elements;
	that.elements = temp;
}

// ================================================================
// Here are several I/O options:
// * <<, >> operators
// * load/store text
// * load/store binary
// * You can also write type-specific (non-template) functions for
//   vector I/O as well, in another source file.
// * I don't know a tidy way to get my favorite formatted I/O options
//   into a template class.  For this reason, there is also
//   ../utils/vector_print.h, with type-specific (non-template) print
//   routines.

// ----------------------------------------------------------------
// I/O format:  elements one per line, delimited by whitespace.
//friend std::ostream & operator<< <>(
//	std::ostream & os,
//	const tvector<element_type> & v);

//friend std::istream & operator>> <>(
//	std::istream & is,
//	tvector<element_type> & v);

// ----------------------------------------------------------------
// Used by tmatrix ostream operator <<.
void row_output(
	std::ostream & os) const
{
	for (int i = 0; i < this->num_elements; i++) {
		if (i > 0)
			os << " ";
		os << this->elements[i];
	}
}

void column_output(
	std::ostream & os) const
{
	for (int i = 0; i < this->num_elements; i++) {
		os << this->elements[i] << "\n";
	}
}

// ----------------------------------------------------------------
void input(
	std::istream & is,
	int read_columns)
{
	const int init_size = 10;
	const int more_size = 10;
	int alloc_size = init_size;

	if (this->elements)
		delete [] this->elements;
	this->elements = 0;
	this->num_elements = 0;

	this->elements = new element_type[alloc_size];
	while (1) {
		if (is.eof()) {
			break;
		}

		if (this->num_elements >= alloc_size) {
			alloc_size += more_size;
			element_type * ptemp = new element_type[alloc_size];
			for (int i = 0; i < this->num_elements; i++)
				ptemp[i] = this->elements[i];
			delete [] this->elements;
			this->elements = ptemp;
		}
		is >> this->elements[this->num_elements];

		// xxx do_columns / do_rows

		if (is.fail()) {
			is.clear();
#if 0
			std::cerr
				<< "tvector istringstream >>: scan failure"
				<< " at vector element " << this->num_elements
				<< ".\n";
#endif
			return;
		}
		this->num_elements++;
	}
}

void column_input(
	std::istream & is)
{
	this->input(is, 1);
}

void row_input(
	std::istream & is)
{
	this->input(is, 0);
}

// ----------------------------------------------------------------
int load_from_text_file(char * input_file_name)
{
	char line[2048];
	std::istream * pis;
	std::ifstream ifs;
	int iscin = 0;

	if (strcmp(input_file_name, "-") == 0) {
		iscin = 1;
		pis = &std::cin;
	}
	else {
		ifs.open(input_file_name, std::ifstream::in);
		if (ifs.fail()) {
			std::cerr << "tvector::load_from_file:  couldn't open \""
				<< input_file_name << "\"\n";
			return 0;
		}
		pis = &ifs;
	}

	const int init_num_elements = 10;
	const int more_num_elements = 10;
	int alloc_num_elements = init_num_elements;

	if (this->elements)
		delete [] this->elements;
	this->elements = new element_type[init_num_elements];
	this->num_elements = 0;

	while (1) {
		if (pis->eof())
			break;

		if (this->num_elements >= alloc_num_elements) {
			alloc_num_elements += more_num_elements;
			element_type * ptemp = new element_type[alloc_num_elements];
			for (int i = 0; i < this->num_elements; i++)
				ptemp[i] = this->elements[i];
			delete [] this->elements;
			this->elements = ptemp;
		}

		// Read a line of text.
		pis->getline(line, sizeof(line));
		if (pis->eof()) {
			pis->clear();
			break;
		}
		else if (pis->fail()) {
			break;
		}

		// Allow comments.
		char * phash = strstr(line, "#");
		if (phash)
			*phash = 0;

		// Allow multiple vectors in the same stream, delimited by
		// carriage returns.
		if (is_whitespace_line(line)) {
			if (this->num_elements == 0)
				continue;
			else
				break;
		}

		// Use element_type's istream >> to scan the line of text as a
		// vector element.
		std::istringstream iss(line, std::ios_base::in);
		iss >> this->elements[this->num_elements];
		if (iss.fail()) {
			pis->setstate(std::ios::failbit);
			std::cerr << "tvector scan failure: \""
				<< line << "\"\n";
			return 0;
		}

		this->num_elements++;
	}

	if (pis->peek() == EOF)
		pis->setstate(std::ios::eofbit);

	if (this->num_elements == 0) {
		pis->setstate(std::ios::badbit);
		std::cerr << "tvector >>:  zero-element vector.\n";
	}

	if (!iscin)
		ifs.close();

	return 1;
}

void load_from_text_file_or_die(char * input_file_name)
{
	if (!this->load_from_text_file(input_file_name)) {
		std::cerr << "tvector:  Couldn't load from \""
			<< input_file_name << "\".\n";
		exit(1);
	}
}

// ----------------------------------------------------------------
int store_to_text_file(char * output_file_name)
{
	if ((strcmp(output_file_name, "-") == 0) || (strcmp(output_file_name, "@") == 0)) {
		std::cout << *this;
		return 1;
	}

	std::ofstream ofs;
	ofs.open(output_file_name, std::ofstream::out);

	if (ofs.fail()) {
		std::cerr << "tvector::store_to_text_file:  couldn't open \""
			<< output_file_name << "\"\n";
		return 0;
	}

	ofs << *this;

	if (ofs.fail()) {
		std::cerr << "tvector::store_to_text_file:  write failure on \""
			<< output_file_name << "\"\n";
		return 0;
	}
	return 1;
}

void store_to_text_file_or_die(char * output_file_name)
{
	if (!this->store_to_text_file(output_file_name)) {
		std::cerr << "tvector:  Couldn't store to \""
			<< output_file_name << "\".\n";
		exit(1);
	}
}

// ----------------------------------------------------------------
int load_from_binary_stream(FILE * fp)
{
	size_t io_size = sizeof(this->elements[0]);
	size_t io_nitems = this->num_elements;
	int rv = 1;
	if (fread((void *)&this->elements[0], io_size, io_nitems, fp) != io_nitems) {
		perror("fread");
		rv = 0;
	}
	return rv;
}

// ----------------------------------------------------------------
int store_to_binary_stream(FILE * fp)
{
	size_t io_size = sizeof(this->elements[0]);
	size_t io_nitems = this->num_elements;
	element_type * temp = new element_type[this->num_elements];
	int rv = 1;
	for (int i = 0; i < this->num_elements; i++)
		temp[i] = this->elements[i];
	if (fwrite((void*)temp, io_size, io_nitems, fp) != io_nitems) {
		perror("fwrite");
		rv = 0;
	}
	delete [] temp;
	return rv;
}

// ----------------------------------------------------------------
int load_from_binary_file(char * input_file_name)
{
	int file_num_bytes;
	struct stat stat_buf;

	if (stat(input_file_name, &stat_buf) != 0) {
		perror("stat");
		fprintf(stderr,
			"tvector::load_from_binary_file: couldn't stat \"%s\".\n",
				input_file_name);
		return 0;
	}

	file_num_bytes =  stat_buf.st_size;
	int element_size = sizeof(element_type);
	if ((file_num_bytes % element_size) != 0) {
		fprintf(stderr, "tvector::load_from_binary_file: element size %d does "
			"not divide file size %d.\n",
			element_size, file_num_bytes);
		exit(1);
	}
	this->num_elements = file_num_bytes / element_size;
	if (this->elements)
		delete [] this->elements;
	this->elements = new element_type[this->num_elements];

	FILE * fp = fopen(input_file_name, "r");
	if (fp == 0) {
		perror("fopen");
		fprintf(stderr,
			"tvector::load_from_binary_file: couldn't open \"%s\".\n",
				input_file_name);
		return 0;
	}
	int rv = this->load_from_binary_stream(fp);
	fclose(fp);
	return rv;
}

void load_from_binary_file_or_die(char * input_file_name)
{
	if (!this->load_from_binary_file(input_file_name)) {
		fprintf(stderr, "Couldn't read \"%s\".\n", input_file_name);
		exit(1);
	}
}

int store_to_binary_file(char * output_file_name)
{
	FILE * fp = fopen(output_file_name, "w");
	if (fp == 0)
		return 0;
	int rv = this->store_to_binary_stream(fp);
	fclose(fp);
	return rv;
}

void store_to_binary_file_or_die(char * output_file_name)
{
	if (!this->store_to_binary_file(output_file_name)) {
		fprintf(stderr, "Couldn't write \"%s\".\n", output_file_name);
		exit(1);
	}
}

// ================================================================
private:
element_type * elements;
int           num_elements;

};

// ================================================================
template <class element_type>
static
std::ostream & operator<< (
	std::ostream & os,
	const tvector<element_type> & v)
{
	v.column_output(os);
	return os;
}

// ----------------------------------------------------------------
template <class element_type>
static
std::istream & operator>> (
	std::istream & is,
	tvector<element_type> & v)
{
	v.column_input(is);
	return is;
}

#endif // TVECTOR_H
