// ================================================================
// Copyright (c) 2004 John Kerl.
// kerl at math dot arizona dot edu
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
// ================================================================

#include <math.h>
#include "gasdev.h"

// ================================================================
// From _Numerical Recipes in C_.
// Returns a normally distributed deviate with zero mean and
// unit variance, using fran32() as the source of uniform deviates.

float float_gasdev(void)
{
	static int iset = 0;
	static float gset;
	float fac, rsq, v1, v2;

	if (iset == 0) {
		// We don't have an extra deviate handy, so pick two uniform
		// numbers in the unit square, repeating until we get a pair
		// in the unit circle.
		do {
			v1 = 2.0 * fran32() - 1.0;
			v2 = 2.0 * fran32() - 1.0;
			rsq = v1*v1 + v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac = sqrt(-2.0 * log(rsq) / rsq);
		// Now make the Box-Muller transformation to get two normal deviates.
		// Return one and save the other for next time.
		gset = v1 * fac;
		iset = 1;
		return v2 * fac;
	}
	else {
		iset = 0;
		return gset;
	}
}

// ----------------------------------------------------------------
float float_gasdev_ms(float mean, float stddev)
{
	return mean + stddev * float_gasdev();
}

// ----------------------------------------------------------------
complex_float complex_float_gasdev(void)
{
	return complex_float_gasdev_ms(0.0, 1.0);
}

// ----------------------------------------------------------------
complex_float complex_float_gasdev_ms(float mean, float stddev)
{
	float mag = float_gasdev_ms(mean, stddev);
	float phz = 2.0 * M_PI * fran32();

	complex_float rv(mag * cos(phz), mag * sin(phz));
	return rv;
}

// ================================================================
float float_gasdev_r(unsigned * pstate0, unsigned * pstate1)
{
	static int iset = 0;
	static float gset;
	float fac, rsq, v1, v2;

	if (iset == 0) {
		// We don't have an extra deviate handy, so pick two uniform
		// numbers in the unit square, repeating until we get a pair
		// in the unit circle.
		do {
			v1 = 2.0 * fran32_r(pstate0, pstate1) - 1.0;
			v2 = 2.0 * fran32_r(pstate0, pstate1) - 1.0;
			rsq = v1*v1 + v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac = sqrt(-2.0 * log(rsq) / rsq);
		// Now make the Box-Muller transformation to get two normal deviates.
		// Return one and save the other for next time.
		gset = v1 * fac;
		iset = 1;
		return v2 * fac;
	}
	else {
		iset = 0;
		return gset;
	}
}

// ----------------------------------------------------------------
float float_gasdev_ms_r(float mean, float stddev,
	unsigned * pstate0, unsigned * pstate1)
{
	return mean + stddev * float_gasdev_r(pstate0, pstate1);
}

// ----------------------------------------------------------------
complex_float complex_float_gasdev_r(unsigned * pstate0, unsigned * pstate1)
{
	return complex_float_gasdev_ms_r(0.0, 1.0, pstate0, pstate1);
}

// ----------------------------------------------------------------
complex_float complex_float_gasdev_ms_r(float mean, float stddev,
	unsigned * pstate0, unsigned * pstate1)
{
	float mag = float_gasdev_ms_r(mean, stddev, pstate0, pstate1);
	float phz = 2.0 * M_PI * fran32_r(pstate0, pstate1);

	complex_float rv(mag * cos(phz), mag * sin(phz));
	return rv;
}

// ================================================================
double double_gasdev(void)
{
	static int iset = 0;
	static double gset;
	double fac, rsq, v1, v2;

	if (iset == 0) {
		// We don't have an extra deviate handy, so pick two uniform
		// numbers in the unit square, repeating until we get a pair
		// in the unit circle.
		do {
			v1 = 2.0 * dran32() - 1.0;
			v2 = 2.0 * dran32() - 1.0;
			rsq = v1*v1 + v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac = sqrt(-2.0 * log(rsq) / rsq);
		// Now make the Box-Muller transformation to get two normal deviates.
		// Return one and save the other for next time.
		gset = v1 * fac;
		iset = 1;
		return v2 * fac;
	}
	else {
		iset = 0;
		return gset;
	}
}

// ----------------------------------------------------------------
double double_gasdev_ms(double mean, double stddev)
{
	return mean + stddev * double_gasdev();
}

// ----------------------------------------------------------------
complex_double complex_double_gasdev(void)
{
	return complex_double_gasdev_ms(0.0, 1.0);
}

// ----------------------------------------------------------------
complex_double complex_double_gasdev_ms(double mean, double stddev)
{
	double mag = double_gasdev_ms(mean, stddev);
	double phz = 2.0 * M_PI * dran32();

	complex_double rv(mag * cos(phz), mag * sin(phz));
	return rv;
}

// ================================================================
double double_gasdev_r(unsigned * pstate0, unsigned * pstate1)
{
	static int iset = 0;
	static double gset;
	double fac, rsq, v1, v2;

	if (iset == 0) {
		// We don't have an extra deviate handy, so pick two uniform
		// numbers in the unit square, repeating until we get a pair
		// in the unit circle.
		do {
			v1 = 2.0 * dran32_r(pstate0, pstate1) - 1.0;
			v2 = 2.0 * dran32_r(pstate0, pstate1) - 1.0;
			rsq = v1*v1 + v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac = sqrt(-2.0 * log(rsq) / rsq);
		// Now make the Box-Muller transformation to get two normal deviates.
		// Return one and save the other for next time.
		gset = v1 * fac;
		iset = 1;
		return v2 * fac;
	}
	else {
		iset = 0;
		return gset;
	}
}

// ----------------------------------------------------------------
double double_gasdev_ms_r(double mean, double stddev,
	unsigned * pstate0, unsigned * pstate1)
{
	return mean + stddev * double_gasdev_r(pstate0, pstate1);
}

// ----------------------------------------------------------------
complex_double complex_double_gasdev_r(unsigned * pstate0, unsigned * pstate1)
{
	return complex_double_gasdev_ms_r(0.0, 1.0, pstate0, pstate1);
}

// ----------------------------------------------------------------
complex_double complex_double_gasdev_ms_r(double mean, double stddev,
	unsigned * pstate0, unsigned * pstate1)
{
	double mag = double_gasdev_ms_r(mean, stddev, pstate0, pstate1);
	double phz = 2.0 * M_PI * dran32_r(pstate0, pstate1);

	complex_double rv(mag * cos(phz), mag * sin(phz));
	return rv;
}
